# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
import math
from decimal import Decimal
from itertools import pairwise
from operator import add, mul, sub, truediv
from random import random
from typing import Any, Iterable

import pytest

import faebryk.library._F as F
from faebryk.core.cpp import Graph
from faebryk.core.module import Module
from faebryk.core.node import Node
from faebryk.core.parameter import (
    Abs,
    Add,
    And,
    Arithmetic,
    ConstrainableExpression,
    Expression,
    GreaterOrEqual,
    GreaterThan,
    Intersection,
    Is,
    IsSubset,
    IsSuperset,
    LessOrEqual,
    Log,
    Max,
    Multiply,
    Not,
    Or,
    Parameter,
    ParameterOperatable,
    Power,
    Round,
    Sin,
    Subtract,
    SymmetricDifference,
    Union,
)
from faebryk.core.solver.defaultsolver import DefaultSolver
from faebryk.core.solver.utils import (
    CanonicalExpression,
    CanonicalLiteral,
    Contradiction,
    ContradictionByLiteral,
)
from faebryk.libs.brightness import TypicalLuminousIntensity
from faebryk.libs.library import L
from faebryk.libs.library.L import DiscreteSet, Range, RangeWithGaps, Single
from faebryk.libs.picker.lcsc import PickedPartLCSC
from faebryk.libs.picker.localpick import PickerOption, pick_module_by_params
from faebryk.libs.picker.picker import pick_part_recursively
from faebryk.libs.sets.quantity_sets import (
    Quantity_Interval_Disjoint,
    Quantity_Set,
)
from faebryk.libs.sets.sets import BoolSet, EnumSet, as_lit
from faebryk.libs.units import P, Quantity, dimensionless, quantity
from faebryk.libs.util import cast_assert, not_none, times
from test.common.resources.fabll_modules.RP2040 import RP2040
from test.common.resources.fabll_modules.RP2040_ReferenceDesign import (
    RP2040_ReferenceDesign,
)
from test.common.resources.fabll_modules.USB_C_PSU_Vertical import USB_C_PSU_Vertical

logger = logging.getLogger(__name__)


def _create_letters(
    n: int, units=dimensionless
) -> tuple[ParameterOperatable.ReprContext, list[Parameter], Graph]:
    context = ParameterOperatable.ReprContext()

    out = []

    class App(Node):
        def __preinit__(self) -> None:
            for _ in range(n):
                p = Parameter(units=units)
                name = p.compact_repr(context)
                self.add(p, name)
                out.append(p)

    app = App()
    return context, out, app.get_graph()


def test_solve_phase_one():
    solver = DefaultSolver()

    def Voltage():
        return L.p_field(units=P.V, within=Range(0 * P.V, 10 * P.kV))

    class App(Module):
        voltage1 = Voltage()
        voltage2 = Voltage()
        voltage3 = Voltage()

    app = App()
    voltage1 = app.voltage1
    voltage2 = app.voltage2
    voltage3 = app.voltage3

    voltage1.alias_is(voltage2)
    voltage3.alias_is(voltage1 + voltage2)

    voltage1.alias_is(Range(1 * P.V, 3 * P.V))
    voltage3.alias_is(Range(2 * P.V, 6 * P.V))

    solver.simplify_symbolically(voltage1.get_graph())


def test_simplify():
    class App(Module):
        ops = L.list_field(
            10,
            lambda: Parameter(
                units=dimensionless, within=Range(0, 1, units=dimensionless)
            ),
        )

    app = App()

    # (((((((((((A + B + 1) + C + 2) * D * 3) * E * 4) * F * 5) * G * (A - A)) + H + 7)
    #  + I + 8) + J + 9) - 3) - 4) < 11
    # => (H + I + J + 17) < 11
    constants: list[ParameterOperatable.NumberLike] = [
        quantity(c, dimensionless) for c in range(0, 10)
    ]
    constants[5] = app.ops[0] - app.ops[0]
    constants[9] = RangeWithGaps(Range(0 * dimensionless, 1 * dimensionless))
    acc = app.ops[0]
    for i, p in enumerate(app.ops[1:3]):
        acc += p + constants[i]
    for i, p in enumerate(app.ops[3:7]):
        acc *= p * constants[i + 3]
    for i, p in enumerate(app.ops[7:]):
        acc += p + constants[i + 7]

    acc = (acc - quantity(3, dimensionless)) - quantity(4, dimensionless)
    assert isinstance(acc, Subtract)
    (acc <= quantity(11, dimensionless)).constrain()

    G = acc.get_graph()
    solver = DefaultSolver()
    solver.simplify_symbolically(G)
    # TODO actually test something


def test_simplify_logic_and():
    class App(Module):
        p = L.list_field(4, lambda: Parameter(domain=L.Domains.BOOL()))

    app = App()
    anded = And(app.p[0], True)
    for p in app.p[1:]:
        anded = anded & p
    anded = anded & anded

    anded.constrain()
    G = anded.get_graph()
    solver = DefaultSolver()
    solver.simplify_symbolically(G)
    # TODO actually test something


def test_shortcircuit_logic_and():
    p0 = Parameter(domain=L.Domains.BOOL())
    expr = p0 & False
    expr.constrain()
    G = expr.get_graph()
    solver = DefaultSolver()

    with pytest.raises(ContradictionByLiteral):
        solver.simplify_symbolically(G)


def test_shortcircuit_logic_or():
    class App(Module):
        p = L.list_field(4, lambda: Parameter(domain=L.Domains.BOOL()))

    app = App()
    ored = Or(app.p[0], True)
    for p in app.p[1:]:
        ored = ored | p
    ored = ored | ored

    ored.constrain()
    G = ored.get_graph()
    solver = DefaultSolver()
    repr_map = solver.simplify_symbolically(G).data.mutation_map
    assert repr_map.try_get_literal(ored) == BoolSet(True)


def test_inequality_to_set():
    p0 = Parameter(units=dimensionless)
    p0.constrain_le(2.0)
    p0.constrain_ge(1.0)
    solver = DefaultSolver()
    solver.update_superset_cache(p0)
    assert solver.inspect_get_known_supersets(p0) == RangeWithGaps((1.0, 2.0))


def test_remove_obvious_tautologies():
    p0, p1, p2 = times(3, Parameter)
    p0.alias_is(p1 + p2)
    p1.constrain_ge(0.0)
    p2.constrain_ge(0.0)
    p2.alias_is(p2)

    G = p0.get_graph()
    solver = DefaultSolver()
    solver.simplify_symbolically(G)
    # TODO actually test something


def test_subset_of_literal():
    p0, p1, p2 = (
        Parameter(units=dimensionless, within=Range(0, i, units=dimensionless))
        for i in range(3)
    )
    p0.alias_is(p1)
    p1.alias_is(p2)

    solver = DefaultSolver()
    solver.update_superset_cache(p0, p1, p2)
    for p in (p0, p1, p2):
        assert solver.inspect_get_known_supersets(p) == RangeWithGaps((0.0, 0.0))


def test_alias_classes():
    A, B, C, D, E = (Parameter() for _ in range(5))
    A.alias_is(B)
    addition = C + D
    B.alias_is(addition)
    addition2 = D + C
    E.alias_is(addition2)

    G = A.get_graph()
    context = ParameterOperatable.ReprContext()
    for p in (A, B, C, D, E):
        p.compact_repr(context)
    solver = DefaultSolver()
    solver.simplify_symbolically(G, print_context=context)
    # TODO actually test something


def test_solve_realworld():
    app = RP2040()
    solver = DefaultSolver()
    solver.simplify_symbolically(app.get_graph())
    # TODO actually test something


@pytest.mark.slow
def test_solve_realworld_bigger():
    app = RP2040_ReferenceDesign()
    F.is_bus_parameter.resolve_bus_parameters(app.get_graph())

    solver = DefaultSolver()
    solver.simplify_symbolically(app.get_graph())
    # TODO actually test something


@pytest.mark.slow
@pytest.mark.usefixtures("setup_project_config")
def test_solve_realworld_biggest():
    class App(Module):
        led = L.f_field(F.LEDIndicator)(use_mosfet=False)
        mcu: RP2040_ReferenceDesign
        usb_power: USB_C_PSU_Vertical

        def __preinit__(self):
            self.led.led.led.color.constrain_subset(F.LED.Color.YELLOW)
            self.led.led.led.brightness.constrain_subset(
                TypicalLuminousIntensity.APPLICATION_LED_INDICATOR_INSIDE.value
            )

            self.usb_power.power_out.connect(self.mcu.usb.usb_if.buspower)
            self.mcu.rp2040.gpio[25].connect(self.led.logic_in)
            self.mcu.rp2040.pinmux.enable(self.mcu.rp2040.gpio[25])

    app = App()
    F.is_bus_parameter.resolve_bus_parameters(app.get_graph())
    solver = DefaultSolver()
    solver.simplify_symbolically(app.get_graph())

    pick_part_recursively(app, solver)


def test_inspect_known_superranges():
    p0 = Parameter(units=P.V, within=Range(1 * P.V, 10 * P.V))
    p0.alias_is(Range(1 * P.V, 3 * P.V) + Range(4 * P.V, 6 * P.V))
    solver = DefaultSolver()
    solver.update_superset_cache(p0)
    assert solver.inspect_get_known_supersets(p0) == RangeWithGaps((5 * P.V, 9 * P.V))


def test_obvious_contradiction_by_literal():
    p0 = Parameter(units=P.V)
    p1 = Parameter(units=P.V)

    p0.alias_is(Range(0 * P.V, 10 * P.V))
    p1.alias_is(Range(5 * P.V, 10 * P.V))

    p0.alias_is(p1)

    G = p0.get_graph()
    solver = DefaultSolver()
    with pytest.raises(ContradictionByLiteral):
        solver.simplify_symbolically(G)


def test_subset_is():
    A, B = params = times(2, lambda: Parameter(domain=L.Domains.Numbers.REAL()))

    A.alias_is(Range(0, 15))
    B.constrain_subset(Range(5, 20))
    A.alias_is(B)

    context = ParameterOperatable.ReprContext()
    for p in params:
        p.compact_repr(context)

    solver = DefaultSolver()
    with pytest.raises(ContradictionByLiteral):
        solver.simplify_symbolically(A.get_graph())


def test_subset_is_expr():
    A, B, C = params = times(3, lambda: Parameter(domain=L.Domains.Numbers.REAL()))

    context = ParameterOperatable.ReprContext()
    for p in params:
        p.compact_repr(context)

    E = A + B
    C.alias_is(Range(0, 15))
    E.constrain_subset(Range(5, 20))

    C.alias_is(E)

    solver = DefaultSolver()
    with pytest.raises(ContradictionByLiteral):
        solver.simplify_symbolically(A.get_graph(), print_context=context)


def test_subset_single_alias():
    A = Parameter(units=P.V)
    A.constrain_subset(Single(1 * P.V))

    solver = DefaultSolver()
    repr_map = solver.simplify_symbolically(A.get_graph()).data.mutation_map
    assert repr_map.try_get_literal(A) == Single(1 * P.V)


def test_very_simple_alias_class():
    A, B, C = params = times(3, lambda: Parameter(units=P.V))
    A.alias_is(B)
    B.alias_is(C)

    context = ParameterOperatable.ReprContext()
    for p in params:
        p.compact_repr(context)

    solver = DefaultSolver()
    repr_map = solver.simplify_symbolically(A.get_graph()).data.mutation_map
    assert (
        repr_map.try_get_literal(A)
        == repr_map.try_get_literal(B)
        == repr_map.try_get_literal(C)
    )


def test_domain():
    p0 = Parameter(units=P.V, within=Range(0 * P.V, 10 * P.V))
    p0.alias_is(Range(15 * P.V, 20 * P.V))

    solver = DefaultSolver()
    with pytest.raises(ContradictionByLiteral):
        solver.simplify_symbolically(p0.get_graph())


def test_less_obvious_contradiction_by_literal():
    A = Parameter(units=P.V)
    B = Parameter(units=P.V)
    C = Parameter(units=P.V)

    A.alias_is(Range(0.0 * P.V, 10.0 * P.V))
    B.alias_is(Range(5.0 * P.V, 10.0 * P.V))
    C.alias_is(A + B)
    C.alias_is(Range(0.0 * P.V, 15.0 * P.V))

    print_context = ParameterOperatable.ReprContext()
    for p in (A, B, C):
        p.compact_repr(print_context)

    G = A.get_graph()
    solver = DefaultSolver()
    with pytest.raises(ContradictionByLiteral):
        solver.simplify_symbolically(G, print_context=print_context)


def test_symmetric_inequality_correlated():
    p0 = Parameter(units=P.V)
    p1 = Parameter(units=P.V)

    p0.alias_is(Range(0 * P.V, 10 * P.V))
    p1.alias_is(p0)

    (p0 >= p1).constrain()
    (p0 <= p1).constrain()

    G = p0.get_graph()
    solver = DefaultSolver()
    repr_map = solver.simplify_symbolically(G).data.mutation_map
    assert repr_map.try_get_literal(p0) == repr_map.try_get_literal(p1)
    assert repr_map.try_get_literal(p0) == Range(0 * P.V, 10 * P.V)


@pytest.mark.parametrize(
    "expr_type, operands, expected",
    [
        (Add, (5, 10), 15),
        # (Subtract, (5, 10), -5),
        # (Multiply, (5, 10), 50),
        # (Divide, (5, 10), 0.5),
    ],
)
def test_simple_literal_folds_arithmetic(
    expr_type: type[Arithmetic], operands: Iterable[float], expected: float
):
    expected_result = Quantity_Interval_Disjoint.from_value(expected)
    used_operands = [Quantity_Interval_Disjoint.from_value(o) for o in operands]

    p0 = Parameter(units=dimensionless)
    p1 = Parameter(units=dimensionless)
    p0.alias_is(used_operands[0])
    p1.alias_is(used_operands[1])

    expr = expr_type(p0, p1)
    (expr <= 100.0).constrain()
    G = expr.get_graph()

    solver = DefaultSolver()
    repr_map = solver.simplify_symbolically(G).data.mutation_map
    deduced_subset = repr_map.try_get_literal(expr, allow_subset=True)
    assert deduced_subset == expected_result


@pytest.mark.parametrize(
    "expr_type, operands, expected",
    [
        (Add, (5, 10), 15),
        (Add, (-5, 15), 10),
        (Add, ((0, 10), 5), (5, 15)),
        (Add, ((0, 10), (-10, 0)), (-10, 10)),
        (Add, (5, 5, 5), 15),
        # (Subtract, (5, 10), -5),
        # (Multiply, (5, 10), 50),
        # (Divide, (5, 10), 0.5),
    ],
)
def test_super_simple_literal_folding(
    expr_type: type[Arithmetic], operands: Iterable[Any], expected: Any
):
    q_operands = [Quantity_Interval_Disjoint.from_value(o) for o in operands]
    expr = expr_type(*q_operands)
    solver = DefaultSolver()

    (expr <= 100.0).constrain()
    G = expr.get_graph()

    repr_map = solver.simplify_symbolically(G).data.mutation_map
    assert repr_map.try_get_literal(expr) == Quantity_Interval_Disjoint.from_value(
        expected
    )


def test_literal_folding_add_multiplicative():
    A = Parameter(units=dimensionless)
    B = Parameter(units=dimensionless)

    expr = A + (A * 2) + (5 * A) + B + (A * B * 2) - B
    # expect: 8A + 2AB

    (expr <= 100.0).constrain()

    G = expr.get_graph()
    solver = DefaultSolver()
    repr_map = solver.simplify_symbolically(G).data.mutation_map

    rep_add = repr_map.map_forward(expr).maps_to
    rep_A = repr_map.map_forward(A).maps_to
    rep_B = repr_map.map_forward(B).maps_to
    assert isinstance(rep_add, Add)
    context = repr_map.output_print_context
    assert len(rep_add.operands) == 2, f"{rep_add.compact_repr(context)}"
    mul1, mul2 = rep_add.operands

    assert isinstance(mul1, Multiply)
    assert isinstance(mul2, Multiply)
    assert any(
        set(m.operands) == {rep_A, Quantity_Interval_Disjoint.from_value(8)}
        for m in (mul1, mul2)
    )
    assert any(
        set(m.operands) == {rep_A, rep_B, Quantity_Interval_Disjoint.from_value(2)}
        for m in (mul1, mul2)
    )


def test_literal_folding_add_multiplicative_2():
    A = Parameter()
    B = Parameter()

    expr = (
        A
        + (A * 2)
        + Quantity_Interval_Disjoint.from_value(10)
        + (5 * A)
        + Quantity_Interval_Disjoint.from_value(0)
        + B
    )
    (expr <= 100.0).constrain()

    G = expr.get_graph()
    solver = DefaultSolver()
    repr_map = solver.simplify_symbolically(G).data.mutation_map
    rep_add = repr_map.map_forward(expr).maps_to
    a_res = repr_map.map_forward(A).maps_to
    b_res = repr_map.map_forward(B).maps_to
    assert isinstance(rep_add, Add)
    assert a_res is not None
    a_ops = [
        op
        for op in a_res.get_operations()
        if isinstance(op, Multiply)
        and Quantity_Interval_Disjoint.from_value(8) in op.operands
    ]
    assert len(a_ops) == 1
    mul = next(iter(a_ops))
    assert set(rep_add.operands) == {
        b_res,
        Quantity_Interval_Disjoint.from_value(10),
        mul,
    }


def test_transitive_subset():
    A = Parameter(domain=L.Domains.Numbers.REAL())
    B = Parameter(domain=L.Domains.Numbers.REAL())
    C = Parameter(domain=L.Domains.Numbers.REAL())

    A.constrain_subset(B)
    B.constrain_subset(C)

    context = ParameterOperatable.ReprContext()
    for p in (A, B, C):
        p.compact_repr(context)

    C.alias_is(Range(0, 10))

    solver = DefaultSolver()
    repr_map = solver.simplify_symbolically(
        A.get_graph(), print_context=context
    ).data.mutation_map
    assert repr_map.try_get_literal(A, allow_subset=True) == Range(0, 10)


def test_nested_additions():
    A = Parameter()
    B = Parameter()
    C = Parameter()
    D = Parameter()

    A.alias_is(Quantity_Interval_Disjoint.from_value(1))
    B.alias_is(Quantity_Interval_Disjoint.from_value(1))
    C.alias_is(A + B)
    D.alias_is(C + A)

    solver = DefaultSolver()
    repr_map = solver.simplify_symbolically(A.get_graph()).data.mutation_map

    assert repr_map.try_get_literal(A) == Quantity_Interval_Disjoint.from_value(1)
    assert repr_map.try_get_literal(B) == Quantity_Interval_Disjoint.from_value(1)
    assert repr_map.try_get_literal(C) == Quantity_Interval_Disjoint.from_value(2)
    assert repr_map.try_get_literal(D) == Quantity_Interval_Disjoint.from_value(3)


def test_combined_add_and_multiply_with_ranges():
    A = Parameter()
    B = Parameter()
    C = Parameter()

    A.alias_is(Range.from_center_rel(1, 0.01))
    B.alias_is(Range.from_center_rel(2, 0.01))
    C.alias_is(2 * A + B)

    solver = DefaultSolver()
    solver.update_superset_cache(A, B, C)
    assert solver.inspect_get_known_supersets(C) == Range.from_center_rel(4, 0.01)


def test_voltage_divider_find_v_out_no_division():
    r_top = Parameter()
    r_bottom = Parameter()
    v_in = Parameter()
    v_out = Parameter()

    v_in.alias_is(Range(9, 10))
    r_top.alias_is(Range(10, 100))
    r_bottom.alias_is(Range(10, 100))
    v_out.alias_is(v_in * r_bottom * ((r_top + r_bottom) ** -1))

    solver = DefaultSolver()

    # dependency problem prevents finding precise solution of [9/11, 100/11]
    # TODO: automatically rearrange expression to match
    # v_out.alias_is(v_in * (1 / (1 + (r_top / r_bottom))))
    solver.update_superset_cache(v_in, v_out, r_top, r_bottom)
    assert solver.inspect_get_known_supersets(v_out) == Range(0.45, 50)


def test_voltage_divider_find_v_out_with_division():
    r_top = Parameter()
    r_bottom = Parameter()
    v_in = Parameter()
    v_out = Parameter()

    v_in.alias_is(Range(9, 10))
    r_top.alias_is(Range(10, 100))
    r_bottom.alias_is(Range(10, 100))
    v_out.alias_is(v_in * r_bottom / (r_top + r_bottom))

    solver = DefaultSolver()
    solver.update_superset_cache(v_in, v_out, r_top, r_bottom)
    assert solver.inspect_get_known_supersets(v_out) == Range(0.45, 50)


def test_voltage_divider_find_v_out_single_variable_occurrences():
    r_top = Parameter()
    r_bottom = Parameter()
    v_in = Parameter()
    v_out = Parameter()

    v_in.alias_is(Range(9, 10))
    r_top.alias_is(Range(10, 100))
    r_bottom.alias_is(Range(10, 100))
    v_out.alias_is(v_in * (1 / (1 + (r_top / r_bottom))))

    solver = DefaultSolver()
    solver.update_superset_cache(v_in, v_out, r_top, r_bottom)
    assert solver.inspect_get_known_supersets(v_out) == Range(9 / 11, 100 / 11)


def test_voltage_divider_find_v_in():
    r_top = Parameter()
    r_bottom = Parameter()
    v_in = Parameter()
    v_out = Parameter()

    v_out.alias_is(Range(9, 10))
    r_top.alias_is(Range(10, 100))
    r_bottom.alias_is(Range(10, 100))
    v_out.alias_is(v_in * r_bottom / (r_top + r_bottom))

    solver = DefaultSolver()

    # TODO: should find [9.9, 100]
    solver.update_superset_cache(v_in, v_out, r_top, r_bottom)
    assert solver.inspect_get_known_supersets(v_in) == Range(1.8, 200)


def test_voltage_divider_find_resistances():
    r_top = Parameter(units=P.ohm)
    r_bottom = Parameter(units=P.ohm)
    v_in = Parameter(units=P.V)
    v_out = Parameter(units=P.V)
    r_total = Parameter(units=P.ohm)

    v_in.alias_is(Range(9 * P.V, 10 * P.V))
    v_out.alias_is(Range(0.9 * P.V, 1 * P.V))
    r_total.alias_is(Quantity_Interval_Disjoint.from_value(100 * P.ohm))
    r_total.alias_is(r_top + r_bottom)
    v_out.alias_is(v_in * r_bottom / (r_top + r_bottom))

    solver = DefaultSolver()
    # FIXME: this test looks funky
    solver.update_superset_cache(v_in, v_out, r_top, r_bottom, r_total)
    assert solver.inspect_get_known_supersets(v_out) == Range(0.9 * P.V, 1 * P.V)

    # TODO: specify r_top (with tolerance), finish solving to find r_bottom


def test_voltage_divider_find_r_top():
    r_top = Parameter(units=P.ohm)
    r_bottom = Parameter(units=P.ohm)
    v_in = Parameter(units=P.V)
    v_out = Parameter(units=P.V)

    v_in.alias_is(Range.from_center_rel(10 * P.V, 0.01))
    v_out.alias_is(Range.from_center_rel(1 * P.V, 0.01))
    r_bottom.alias_is(Range.from_center_rel(1 * P.ohm, 0.01))
    v_out.alias_is(v_in * r_bottom / (r_top + r_bottom))
    # r_top = (v_in * r_bottom) / v_out - r_bottom

    solver = DefaultSolver()
    solver.update_superset_cache(v_in, v_out, r_top, r_bottom)
    assert solver.inspect_get_known_supersets(r_top) == Range(
        (10 * 0.99**2) / 1.01 - 1.01, (10 * 1.01**2) / 0.99 - 0.99
    )


def test_voltage_divider_reject_invalid_r_top():
    r_top = Parameter(units=P.ohm)
    r_bottom = Parameter(units=P.ohm)
    v_in = Parameter(units=P.V)
    v_out = Parameter(units=P.V)

    v_in.alias_is(Range.from_center_rel(10 * P.V, 0.01))
    v_out.alias_is(Range.from_center_rel(1 * P.V, 0.01))
    v_out.alias_is(v_in * r_bottom / (r_top + r_bottom))

    r_bottom.alias_is(Range.from_center_rel(1 * P.ohm, 0.01))
    r_top.alias_is(Range.from_center_rel(999 * P.ohm, 0.01))

    solver = DefaultSolver()
    with pytest.raises(ContradictionByLiteral):
        solver.simplify_symbolically(r_top.get_graph())


def test_base_unit_switch():
    A = Parameter(units=P.mAh)
    A.alias_is(Range(100 * P.mAh, 600 * P.mAh))
    (A >= 100 * P.mAh).constrain()

    G = A.get_graph()
    solver = DefaultSolver()
    repr_map = solver.simplify_symbolically(G).data.mutation_map
    assert repr_map.try_get_literal(A) == RangeWithGaps.from_value(
        (100 * P.mAh, 600 * P.mAh)
    )


@pytest.mark.parametrize("predicate_type", [Is, IsSubset])
def test_try_fulfill_super_basic(predicate_type: type[ConstrainableExpression]):
    p0 = Parameter(units=P.V)
    p0.alias_is(Range(0 * P.V, 10 * P.V))

    solver = DefaultSolver()
    pred = predicate_type(p0, Range(0 * P.V, 10 * P.V))
    assert solver.try_fulfill(pred, lock=False)


def test_congruence_filter():
    A = Parameter(domain=L.Domains.ENUM(F.LED.Color))
    x = Is(A, EnumSet(F.LED.Color.EMERALD))

    y1 = Not(x).constrain()
    y2 = Not(x).constrain()
    assert y1.is_congruent_to(y2)

    solver = DefaultSolver()
    repr_map = solver.simplify_symbolically(x.get_graph()).data.mutation_map
    assert repr_map.map_forward(y1).maps_to == repr_map.map_forward(y2).maps_to


def test_inspect_enum_simple():
    A = Parameter(domain=L.Domains.ENUM(F.LED.Color))

    A.constrain_subset(F.LED.Color.EMERALD)

    solver = DefaultSolver()
    solver.update_superset_cache(A)
    assert solver.inspect_get_known_supersets(A) == F.LED.Color.EMERALD


def test_regression_enum_contradiction():
    A = Parameter(domain=L.Domains.ENUM(F.LED.Color))

    A.constrain_subset(L.EnumSet(F.LED.Color.BLUE, F.LED.Color.RED))

    solver = DefaultSolver()
    with pytest.raises(Contradiction):
        solver.try_fulfill(Is(A, F.LED.Color.EMERALD), lock=False)


def test_inspect_enum_led():
    led = F.LED()

    led.color.constrain_subset(F.LED.Color.EMERALD)

    solver = DefaultSolver()
    solver.update_superset_cache(led.color)
    assert solver.inspect_get_known_supersets(led.color) == F.LED.Color.EMERALD


@pytest.mark.usefixtures("setup_project_config")
def test_simple_pick():
    led = F.LED()

    solver = DefaultSolver()
    pick_module_by_params(
        led,
        solver,
        [
            PickerOption(
                part=PickedPartLCSC(
                    manufacturer="Everlight Elec",
                    partno="19-217/GHC-YR1S2/3T",
                    supplier_partno="C72043",
                ),
                params={
                    "color": L.EnumSet(F.LED.Color.EMERALD),
                    "max_brightness": 285 * P.mcandela,
                    "forward_voltage": L.Single(3.7 * P.volt),
                    "max_current": 100 * P.mA,
                },  # type: ignore
                pinmap={"1": led.cathode, "2": led.anode},
            ),
        ],
    )

    assert led.has_trait(F.has_part_picked)
    assert (
        cast_assert(PickedPartLCSC, led.get_trait(F.has_part_picked).get_part()).lcsc_id
        == "C72043"
    )


@pytest.mark.usefixtures("setup_project_config")
def test_simple_negative_pick():
    led = F.LED()
    led.color.constrain_subset(L.EnumSet(F.LED.Color.RED, F.LED.Color.BLUE))

    solver = DefaultSolver()
    pick_module_by_params(
        led,
        solver,
        [
            PickerOption(
                part=PickedPartLCSC(
                    manufacturer="Everlight Elec",
                    partno="19-217/GHC-YR1S2/3T",
                    supplier_partno="C72043",
                ),
                params={
                    "color": L.EnumSet(F.LED.Color.EMERALD),
                    "max_brightness": 285 * P.mcandela,
                    "forward_voltage": L.Single(3.7 * P.volt),
                    "max_current": 100 * P.mA,
                },  # type: ignore
                pinmap={"1": led.cathode, "2": led.anode},
            ),
            PickerOption(
                part=PickedPartLCSC(
                    manufacturer="Everlight Elec",
                    partno="19-217/BHC-ZL1M2RY/3T",
                    supplier_partno="C72041",
                ),
                params={
                    "color": L.EnumSet(F.LED.Color.BLUE),
                    "max_brightness": 28.5 * P.mcandela,
                    "forward_voltage": L.Single(3.1 * P.volt),
                    "max_current": 100 * P.mA,
                },  # type: ignore
                pinmap={"1": led.cathode, "2": led.anode},
            ),
        ],
    )

    assert led.has_trait(F.has_part_picked)
    assert (
        cast_assert(PickedPartLCSC, led.get_trait(F.has_part_picked).get_part()).lcsc_id
        == "C72041"
    )


def test_jlcpcb_pick_resistor():
    resistor = F.Resistor()
    resistor.resistance.constrain_subset(L.Range(10 * P.ohm, 100 * P.ohm))

    solver = DefaultSolver()
    pick_part_recursively(resistor, solver)

    assert resistor.has_trait(F.has_part_picked)
    print(resistor.get_trait(F.has_part_picked).get_part())


def test_jlcpcb_pick_capacitor():
    capacitor = F.Capacitor()
    capacitor.capacitance.constrain_subset(L.Range(100 * P.nF, 1 * P.uF))
    capacitor.max_voltage.constrain_ge(50 * P.V)

    solver = DefaultSolver()
    pick_part_recursively(capacitor, solver)

    assert capacitor.has_trait(F.has_part_picked)
    print(capacitor.get_trait(F.has_part_picked).get_part())


@pytest.mark.xfail(reason="TODO: add support for leds")
def test_jlcpcb_pick_led():
    led = F.LED()
    led.color.constrain_subset(L.EnumSet(F.LED.Color.EMERALD))
    led.max_current.constrain_ge(10 * P.mA)

    solver = DefaultSolver()
    pick_part_recursively(led, solver)

    assert led.has_trait(F.has_part_picked)
    print(led.get_trait(F.has_part_picked).get_part())


@pytest.mark.xfail(reason="TODO: add support for powered leds")
def test_jlcpcb_pick_powered_led_simple():
    led = F.PoweredLED()
    led.led.color.constrain_subset(L.EnumSet(F.LED.Color.EMERALD))
    led.power.voltage.constrain_subset(L.Range(1.8 * P.V, 5.5 * P.V))
    led.led.forward_voltage.constrain_subset(L.Range(1 * P.V, 4 * P.V))

    solver = DefaultSolver()
    children_mods = led.get_children_modules(direct_only=False, types=(Module,))

    pick_part_recursively(led, solver)

    picked_parts = [mod for mod in children_mods if mod.has_trait(F.has_part_picked)]
    assert len(picked_parts) == 2
    print([(p, p.get_trait(F.has_part_picked).get_part()) for p in picked_parts])


@pytest.mark.xfail(reason="TODO: add support for powered leds")
def test_jlcpcb_pick_powered_led_regression():
    led = F.PoweredLED()
    led.led.color.constrain_subset(F.LED.Color.RED)
    led.power.voltage.alias_is(3 * P.V)
    led.led.brightness.constrain_subset(
        TypicalLuminousIntensity.APPLICATION_LED_INDICATOR_INSIDE.value
    )

    solver = DefaultSolver()
    children_mods = led.get_children_modules(direct_only=False, types=(Module,))

    pick_part_recursively(led, solver)

    picked_parts = [mod for mod in children_mods if mod.has_trait(F.has_part_picked)]
    assert len(picked_parts) == 2
    for p in picked_parts:
        print(p.get_full_name(types=False), p.get_trait(F.has_part_picked).get_part())
        print(p.pretty_params(solver))


@pytest.mark.parametrize(
    "op, x_op_y, y, x_expected",
    [
        (
            Add,
            Range.from_center_rel(3, 0.01),
            Range.from_center_rel(1, 0.01),
            Range.from_center_rel(2, 0.02),
        )
    ],
)
def test_simple_parameter_isolation(
    op: type[Arithmetic], x_op_y: Range, y: Range, x_expected: Range
):
    X = Parameter()
    Y = Parameter()

    op(X, Y).alias_is(x_op_y)
    Y.alias_is(y)

    solver = DefaultSolver()
    solver.update_superset_cache(X, Y)

    assert solver.inspect_get_known_supersets(X) == x_expected


def test_abstract_lowpass():
    Li = Parameter(units=P.H)
    C = Parameter(units=P.F)
    fc = Parameter(units=P.Hz)

    # formula
    fc.alias_is(1 / (2 * math.pi * (C * Li).operation_sqrt()))

    # input
    Li.alias_is(Range.from_center_rel(1 * P.uH, 0.01))
    fc.alias_is(Range.from_center_rel(1000 * P.Hz, 0.01))

    # solve
    solver = DefaultSolver()
    solver.update_superset_cache(Li, C, fc)

    assert solver.inspect_get_known_supersets(C) == Range(
        6.158765796 * P.GF, 6.410118344 * P.GF
    )


def test_param_isolation():
    X = Parameter()
    Y = Parameter()

    (X + Y).alias_is(Range.from_center_rel(3, 0.01))
    Y.alias_is(Range.from_center_rel(1, 0.01))

    solver = DefaultSolver()
    solver.update_superset_cache(X, Y)

    assert solver.inspect_get_known_supersets(X) == Range.from_center_rel(2, 0.02)


@pytest.mark.parametrize(
    "op",
    [
        add,
        mul,
        sub,
        truediv,
    ],
)
def test_extracted_literal_folding(op):
    A = Parameter()
    B = Parameter()
    C = Parameter(domain=L.Domains.Numbers.REAL())

    lit1 = Range(0, 10)
    lit2 = Range(10, 20)
    lito = op(lit1, lit2)

    A.alias_is(lit1)
    B.alias_is(lit2)

    op(A, B).alias_is(C)

    solver = DefaultSolver()
    solver.update_superset_cache(A, B, C)

    assert solver.inspect_get_known_supersets(C) == lito


def test_fold_pow():
    A = Parameter()
    B = Parameter()

    lit = RangeWithGaps(Range(5, 6))
    lit_operand = 2

    A.alias_is(lit)
    B.alias_is(A**lit_operand)

    solver = DefaultSolver()
    repr_map = solver.simplify_symbolically(B.get_graph()).data.mutation_map

    res = repr_map.try_get_literal(B)
    assert res == lit**lit_operand


def test_graph_split():
    class App(Module):
        A: Parameter
        B: Parameter

    app = App()

    C = Parameter()
    D = Parameter()
    app.A.alias_is(C)
    app.B.alias_is(D)

    context = ParameterOperatable.ReprContext()
    for p in (app.A, app.B, C, D):
        p.compact_repr(context)

    solver = DefaultSolver()
    repr_map = solver.simplify_symbolically(
        app.get_graph(), print_context=context
    ).data.mutation_map

    assert (
        not_none(repr_map.map_forward(app.A).maps_to).get_graph()
        is not not_none(repr_map.map_forward(app.B).maps_to).get_graph()
    )


def test_ss_single_into_alias():
    A = Parameter()
    B = Parameter()

    A.alias_is(Range(5, 10))
    B.operation_is_subset(5).constrain()
    C = A + B

    solver = DefaultSolver()
    repr_map = solver.simplify_symbolically(C.get_graph()).data.mutation_map

    assert repr_map.try_get_literal(B) == 5
    assert repr_map.try_get_literal(A) == Range(5, 10)


@pytest.mark.parametrize(
    "op, invert",
    [
        (GreaterOrEqual, False),
        (LessOrEqual, True),
        (Is, True),
        (Is, False),
        (IsSubset, True),
        (IsSubset, False),
        (IsSuperset, True),
        (IsSuperset, False),
    ],
)
def test_find_contradiction_by_predicate(op, invert):
    """
    A > B, A is [0, 10], B is [20, 30], A further uncorrelated B
    -> [0,10] > [20, 30]
    """

    A = Parameter()
    B = Parameter()

    A.alias_is(Range(0, 10))
    B.alias_is(Range(20, 30))

    if invert:
        op(B, A).constrain()
    else:
        op(A, B).constrain()

    solver = DefaultSolver()

    with pytest.raises(Contradiction):
        solver.simplify_symbolically(A.get_graph())


def test_find_contradiction_by_gt():
    A = Parameter()
    B = Parameter()

    A.alias_is(Range(0, 10))
    B.alias_is(Range(20, 30))

    (A > B).constrain()

    solver = DefaultSolver()
    with pytest.raises(ContradictionByLiteral):
        solver.simplify_symbolically(A.get_graph())


def test_can_add_parameters():
    A = Parameter()
    B = Parameter()
    C = Parameter()

    A.alias_is(Range(10, 100))
    B.alias_is(Range(10, 100))
    C.alias_is((A + B))

    solver = DefaultSolver()
    solver.update_superset_cache(A, B, C)

    assert solver.inspect_get_known_supersets(C) == Range(20, 200)


def test_ss_estimation_ge():
    A = Parameter()
    B = Parameter()

    A.operation_is_subset(Range(0, 10)).constrain()
    (B >= A).constrain()

    solver = DefaultSolver()
    solver.simplify_symbolically(B.get_graph())


def test_fold_mul_zero():
    A = Parameter()
    B = Parameter()
    C = Parameter()

    A.alias_is(0)
    B.alias_is(Range(10, 20))

    (A * B).alias_is(C)

    solver = DefaultSolver()
    solver.update_superset_cache(A, B, C)

    assert solver.inspect_get_known_supersets(C) == 0


def test_fold_or_true():
    A = Parameter(domain=L.Domains.BOOL())
    B = Parameter(domain=L.Domains.BOOL())
    C = Parameter(domain=L.Domains.BOOL())

    A.alias_is(True)

    (A | B).alias_is(C)

    solver = DefaultSolver()
    solver.update_superset_cache(A, B, C)
    assert solver.inspect_get_known_supersets(C) == BoolSet(True)


def test_fold_not():
    A = Parameter(domain=L.Domains.BOOL())
    B = Parameter(domain=L.Domains.BOOL())

    A.alias_is(False)
    (Not(A)).alias_is(B)

    solver = DefaultSolver()
    solver.update_superset_cache(A, B)
    assert solver.inspect_get_known_supersets(B) == BoolSet(True)


def test_fold_ss_transitive():
    A = Parameter()
    B = Parameter()
    C = Parameter()

    C.operation_is_subset(Range(0, 10)).constrain()
    B.operation_is_subset(C).constrain()
    A.operation_is_subset(B).constrain()

    solver = DefaultSolver()
    solver.update_superset_cache(A, B, C)
    assert solver.inspect_get_known_supersets(A) == Range(0, 10)


def test_ss_intersect():
    A = Parameter()
    B = Parameter()
    C = Parameter()

    A.alias_is(Range(0, 15))
    B.alias_is(Range(10, 20))
    C.constrain_subset(A)
    C.constrain_subset(B)

    solver = DefaultSolver()
    solver.update_superset_cache(A, B, C)
    assert solver.inspect_get_known_supersets(C) == Range(10, 15)


@pytest.mark.parametrize(
    "left, right, expected",
    [
        (
            [Range(0, 10)],
            [Range(0, 10)],
            (True, False),
        ),
        (
            [Range(0, 10)],
            [Range(10, 20)],
            (False, False),
        ),
        (
            [Add(Range(0, 10), Range(0, 20))],
            [Add(Range(0, 10), Range(0, 20))],
            (True, False),
        ),
        (
            [Add(Range(0, 10), Range(0, 20))],
            [Add(Range(0, 20), Range(0, 10))],
            (True, False),
        ),
        (
            [Not(BoolSet(True))],
            [Not(BoolSet(True))],
            (True, True),
        ),
        (
            [Not(Not(BoolSet(True)))],
            [Not(Not(BoolSet(True)))],
            (True, True),
        ),
        (
            [Multiply(Range(0, 10), Range(0, 10))],
            [Multiply(Range(0, 10), Range(0, 10))],
            (True, False),
        ),
        (
            [Multiply(Range(0, math.inf), Range(0, math.inf), Range(0, math.inf))],
            [Multiply(Range(0, math.inf), Range(0, math.inf))],
            (False, False),
        ),
        (
            [Add(Range(0, math.inf), Range(0, math.inf))],
            [Add(Range(0, math.inf))],
            (False, False),
        ),
    ],
)
def test_congruence_lits(left, right, expected):
    assert (
        Expression.are_pos_congruent(left, right, allow_uncorrelated=True)
        == expected[0]
    )
    assert Expression.are_pos_congruent(left, right) == expected[1]


def test_fold_literals():
    A = Parameter()
    A.alias_is(Add(Range(0, 10), Range(0, 10)))

    solver = DefaultSolver()
    solver.update_superset_cache(A)
    assert solver.inspect_get_known_supersets(A) == Range(0, 20)


def test_deduce_negative():
    A = Parameter(domain=L.Domains.BOOL())

    p = Not(A)

    solver = DefaultSolver()
    assert solver.try_fulfill(p, lock=False)


def test_empty_and():
    solver = DefaultSolver()

    p = And()
    assert solver.try_fulfill(p, lock=False)


def test_implication():
    A = Parameter()
    B = Parameter()

    A.constrain_subset(DiscreteSet(5, 10))

    A.operation_is_subset(Single(5)).operation_implies(
        B.operation_is_subset(Range.from_center_rel(100, 0.1))
    ).constrain()
    A.operation_is_subset(Single(10)).operation_implies(
        B.operation_is_subset(Range.from_center_rel(500, 0.1))
    ).constrain()

    A.constrain_subset(Single(10))

    solver = DefaultSolver()
    solver.update_superset_cache(A, B)
    assert solver.inspect_get_known_supersets(B) == Range.from_center_rel(500, 0.1)


@pytest.mark.parametrize("A_value", [5, 10, 15])
def test_mapping(A_value):
    A = Parameter()
    B = Parameter()

    X = Range.from_center_rel(100, 0.1)
    Y = Range.from_center_rel(200, 0.1)
    Z = Range.from_center_rel(300, 0.1)

    mapping = {5: X, 10: Y, 15: Z}
    A.constrain_mapping(B, mapping)  # type: ignore

    A.constrain_subset(A_value)

    solver = DefaultSolver()
    solver.update_superset_cache(A, B)
    assert solver.inspect_get_known_supersets(B) == mapping[A_value]


@pytest.mark.parametrize("op", [Subtract, sub, Add, add])
def test_subtract_zero(op):
    from faebryk.core.solver.utils import make_lit

    A = Parameter()
    A.alias_is(op(make_lit(1), make_lit(0)))

    solver = DefaultSolver()
    solver.update_superset_cache(A)
    assert solver.inspect_get_known_supersets(A) == make_lit(1)


def test_canonical_subtract_zero():
    from faebryk.core.solver.utils import make_lit

    A = Parameter()
    A.alias_is(Multiply(make_lit(0), make_lit(-1)))

    B = Parameter()
    B.alias_is(Add(make_lit(1), Multiply(make_lit(0), make_lit(-1))))

    solver = DefaultSolver()
    solver.update_superset_cache(A, B)
    assert solver.inspect_get_known_supersets(A) == make_lit(0)
    assert solver.inspect_get_known_supersets(B) == make_lit(1)


def test_nested_fold_scalar():
    from faebryk.core.solver.utils import make_lit

    A = Parameter()
    A.alias_is(Add(make_lit(1), Multiply(make_lit(2), make_lit(3))))

    solver = DefaultSolver()
    solver.update_superset_cache(A)
    assert solver.inspect_get_known_supersets(A) == make_lit(7)


def test_regression_lit_mul_fold_powers():
    A = Parameter()
    A.alias_is(Power(2, -1) * Power(2, 0.5))

    solver = DefaultSolver()
    solver.update_superset_cache(A)
    assert solver.inspect_get_known_supersets(A) == 2**-0.5


def test_nested_fold_interval():
    A = Parameter()
    A.alias_is(
        Add(
            Range.from_center_rel(1, 0.1),
            Multiply(Range.from_center_rel(2, 0.1), Range.from_center_rel(3, 0.1)),
        )
    )

    solver = DefaultSolver()
    solver.update_superset_cache(A)
    assert solver.inspect_get_known_supersets(A) == Range(5.76, 8.36)


def test_simplify_non_terminal_manual_test_1():
    """
    Test that non-terminal simplification works
    No assertions, run with
    FBRK_LOG_PICK_SOLVE=y FBRK_SLOG=y and read log
    """

    A = Parameter(units=P.V)
    E = A + A

    solver = DefaultSolver()
    solver.simplify(E)
    E2 = E + A
    A.alias_is(Range(0 * P.V, 10 * P.V))

    solver.simplify(E2)

    solver.simplify_symbolically(E2, terminal=True)

    solver.simplify(E2)


def test_simplify_non_terminal_manual_test_2():
    """
    Test that non-terminal simplification works
    No assertions, run with
    FBRK_LOG_PICK_SOLVE=y FBRK_SLOG=y and read log
    """

    context, ps, graph = _create_letters(3, units=P.V)
    A, B, C = ps

    INCREASE = 20 * P.percent
    TOLERANCE = 5 * P.percent
    increase = as_lit(
        L.Range.from_center_rel(INCREASE, TOLERANCE) + L.Single(100 * P.percent)
    )
    for p1, p2 in pairwise(ps):
        p2.constrain_subset(p1 * increase)
        p1.constrain_subset(p2 / increase)

    solver = DefaultSolver()
    solver.simplify_symbolically(A, terminal=False, print_context=context)

    origin = 1, as_lit(Range(9 * P.V, 11 * P.V))
    ps[origin[0]].alias_is(origin[1])
    solver.simplify(A)

    solver.update_superset_cache(*ps)
    for i, p in enumerate(ps):
        # _inc = increase ** (i - origin[0])
        _inc = 1
        _i = i - origin[0]
        for _ in range(abs(_i)):
            if _i > 0:
                _inc *= increase
            else:
                _inc /= increase

        p_lit = solver.inspect_get_known_supersets(p)
        print(f"{p.compact_repr(context)}, lit:", p_lit)
        assert p_lit.is_subset_of(origin[1] * _inc)
        p.alias_is(p_lit)
        solver.simplify(p)


# XFAIL --------------------------------------------------------------------------------


# extra formula
# C.alias_is(1 / (4 * math.pi**2 * Li * fc**2))
# TODO test with only fc given
@pytest.mark.xfail(reason="Need more powerful expression reordering")  # TODO
def test_abstract_lowpass_ss():
    Li = Parameter(units=P.H)
    C = Parameter(units=P.F)
    fc = Parameter(units=P.Hz)

    # formula
    fc.alias_is(1 / (2 * math.pi * (C * Li).operation_sqrt()))

    # input
    Li_const = RangeWithGaps(Range.from_center_rel(1 * P.uH, 0.01))
    fc_const = RangeWithGaps(Range.from_center_rel(1000 * P.Hz, 0.01))
    Li.constrain_subset(Li_const)
    fc.constrain_subset(fc_const)

    # solve
    solver = DefaultSolver()
    solver.simplify_symbolically(fc.get_graph())

    C_expected = 1 / (4 * math.pi**2 * Li_const * fc_const**2)

    solver.update_superset_cache(Li, C, fc)
    assert solver.inspect_get_known_supersets(C) == C_expected


@pytest.mark.xfail(reason="Need more powerful expression reordering")  # TODO
def test_voltage_divider_find_r_bottom():
    r_top = Parameter(units=P.ohm)
    r_bottom = Parameter(units=P.ohm)
    v_in = Parameter(units=P.V)
    v_out = Parameter(units=P.V)

    # formula
    v_out.alias_is(v_in * r_bottom / (r_top + r_bottom))

    # input
    v_in.alias_is(Range.from_center_rel(10 * P.V, 0.01))
    v_out.alias_is(Range.from_center_rel(1 * P.V, 0.01))
    r_top.alias_is(Range.from_center_rel(9 * P.ohm, 0.01))

    solver = DefaultSolver()
    solver.update_superset_cache(v_in, v_out, r_top)
    assert solver.inspect_get_known_supersets(r_bottom) == Range.from_center_rel(
        1 * P.ohm, 0.01
    )


@pytest.mark.xfail(reason="TODO reenable ge fold")
def test_min_max_single():
    p0 = Parameter(units=P.V)
    p0.alias_is(L.Range(0 * P.V, 10 * P.V))

    p1 = Parameter(units=P.V)
    p1.alias_is(Max(p0))

    solver = DefaultSolver()
    solver.update_superset_cache(p0, p1)
    out = solver.inspect_get_known_supersets(p1)
    assert out == L.Single(10 * P.V)


@pytest.mark.xfail(reason="TODO")
def test_min_max_multi():
    p0 = Parameter(units=P.V)
    p0.alias_is(L.Range(0 * P.V, 10 * P.V))
    p3 = Parameter(units=P.V)
    p3.alias_is(L.Range(4 * P.V, 15 * P.V))

    p1 = Parameter(units=P.V)
    p1.alias_is(Max(p0, p3))

    solver = DefaultSolver()
    solver.update_superset_cache(p0, p1, p3)
    out = solver.inspect_get_known_supersets(p1)
    assert out == L.Single(15 * P.V)


@pytest.mark.xfail(
    reason="Behaviour not implemented https://github.com/atopile/atopile/issues/615"
)
def test_symmetric_inequality_uncorrelated():
    p0 = Parameter(units=P.V)
    p1 = Parameter(units=P.V)

    p0.alias_is(Range(0 * P.V, 10 * P.V))

    (p0 >= p1).constrain()
    (p0 <= p1).constrain()

    # This would only work if p0 is alias p1
    # but we never do implicit alias, because that's very dangerous
    # so this has to throw

    # strategy: if this kind of unequality exists, check if there is an alias
    # and if not, throw

    G = p0.get_graph()
    solver = DefaultSolver()

    with pytest.raises(Contradiction):
        solver.simplify_symbolically(G)


def test_fold_correlated():
    """
    ```
    A is [5, 10], B is [10, 15]
    B is A + 5
    B - A | [10, 15] - [5, 10] = [0, 10] BUT SHOULD BE 5
    ```

    A and B correlated, thus B - A should do ss not alias
    """

    A = Parameter()
    B = Parameter()
    C = Parameter()

    op = add
    op_inv = sub

    lit1 = Range(5, 10)
    lit_operand = Single(5)
    lit2 = op(lit1, lit_operand)

    A.alias_is(lit1)  # A is [5,10]
    B.alias_is(lit2)  # B is [10,15]
    # correlate A and B
    B.alias_is(op(A, lit_operand))  # B is A + 5
    C.alias_is(op_inv(B, A))  # C is B - A

    context = ParameterOperatable.ReprContext()
    for p in (A, B, C):
        p.compact_repr(context)

    solver = DefaultSolver()
    repr_map = solver.simplify_symbolically(
        C.get_graph(), print_context=context
    ).data.mutation_map

    is_lit = repr_map.try_get_literal(C, allow_subset=False)
    ss_lit = repr_map.try_get_literal(C, allow_subset=True)
    assert ss_lit is not None

    # Test for ss estimation
    assert ss_lit.is_subset_of(op_inv(lit2, lit1))  # C ss [10, 15] - 5 == [5, 10]
    # Test for not wrongful is estimation
    assert is_lit != op_inv(lit2, lit1)  # C not is [5, 10]

    # Test for correct is estimation
    try:
        assert is_lit == lit_operand  # C is 5
    except AssertionError:
        pytest.xfail("TODO")


@pytest.mark.parametrize(
    "op,lits,expected",
    [
        (Add, [], 0),
        (Add, [1, 2, 3, 4, 5], 15),
        (Multiply, [1, 2, 3, 4, 5], 120),
        (Multiply, [], 1),
        (Power, [2, 3], 8),
        (Round, [2.4], 2),
        (Round, [Range(-2.6, 5.3)], Range(-3, 5)),
        (Abs, [-2], 2),
        (Abs, [Range(-2, 3)], Range(0, 3)),
        (Sin, [0], 0),
        (Sin, [Range(0, 2 * math.pi)], Range(-1, 1)),
        (Log, [10], math.log(10)),
        (Log, [Range(1, 10)], Range(math.log(1), math.log(10))),
        (Or, [False, False, True], True),
        (Or, [False, BoolSet(True, False), True], True),
        (Or, [], False),
        (Not, [False], True),
        (Intersection, [Range(0, 10), Range(10, 20)], Range(10, 10)),
        (Union, [Range(0, 10), Range(10, 20)], Range(0, 20)),
        (SymmetricDifference, [Range(0, 10), Range(10, 20)], Range(0, 20)),
        (
            SymmetricDifference,
            [Range(0, 10), Range(5, 20)],
            RangeWithGaps(Range(0, 5), Range(10, 20)),
        ),
        (Is, [Range(0, 10), Range(0, 10)], True),
        (GreaterOrEqual, [Range(10, 20), Range(0, 10)], True),
        (GreaterOrEqual, [Range(5, 20), Range(0, 10)], BoolSet(True, False)),
        (GreaterThan, [Range(10, 20), Range(0, 10)], BoolSet(True, False)),
        (GreaterThan, [Range(0, 10), Range(10, 20)], False),
        (IsSubset, [Range(0, 10), Range(0, 20)], True),
    ],
)
def test_exec_pure_literal_expressions(op: type[CanonicalExpression], lits, expected):
    from faebryk.core.solver.symbolic.pure_literal import (
        _exec_pure_literal_expressions,
    )
    from faebryk.core.solver.utils import make_lit

    lits_converted = list(map(make_lit, lits))
    expected_converted = make_lit(expected)

    expr = op(*lits_converted)  # type: ignore
    assert _exec_pure_literal_expressions(expr) == expected_converted

    if op is GreaterThan:
        pytest.xfail("GreaterThan is not supported in solver")

    def _get_param_from_lit(lit: CanonicalLiteral):
        if isinstance(lit, BoolSet):
            p = Parameter(domain=L.Domains.BOOL())
        elif isinstance(lit, Quantity_Set):
            p = Parameter(domain=L.Domains.Numbers.REAL())
        else:
            raise NotImplementedError()
        return p

    result = _get_param_from_lit(expected_converted)
    result.alias_is(expr)

    solver = DefaultSolver()
    repr_map = solver.simplify_symbolically(result.get_graph()).data.mutation_map
    assert repr_map.try_get_literal(result) == expected_converted


@pytest.mark.slow
# @pytest.mark.parametrize(
#    "v_in, v_out, total_current",
#    [
#        (
#            Range(9.9 * P.V, 10.1 * P.V),
#            Range(3.0 * P.V, 3.2 * P.V),
#            Range(1 * P.mA, 3 * P.mA),
#        ),
#    ],
# )
# def test_solve_voltage_divider_complex(v_in, v_out, total_current):
def test_solve_voltage_divider_complex():
    v_in, v_out, total_current = (
        as_lit(Range(9.9 * P.V, 10.1 * P.V)),
        as_lit(Range(3.0 * P.V, 3.2 * P.V)),
        as_lit(Range(1 * P.mA, 3 * P.mA)),
    )

    rdiv = F.ResistorVoltageDivider()

    rdiv.v_in.alias_is(v_in)
    rdiv.v_out.constrain_subset(v_out)
    rdiv.max_current.constrain_subset(total_current)

    # Solve for r_top
    print("Solving for r_top")
    solver = DefaultSolver()
    solver.update_superset_cache(rdiv)

    r_top = solver.inspect_get_known_supersets(rdiv.r_top.resistance)
    assert isinstance(r_top, Quantity_Interval_Disjoint)
    print(f"r_top: {r_top}")
    expected_r_top = (v_in - v_out) / total_current
    print(f"Expected r_top: {expected_r_top}")
    assert r_top == expected_r_top

    # Pick a random valid resistor for r_top
    rand_ = Decimal(random())
    r_any_nominal = r_top.min_elem + rand_ * (r_top.max_elem - r_top.min_elem)
    assert isinstance(r_any_nominal, Quantity)
    r_any = L.Range.from_center_rel(r_any_nominal, 0.01)
    rdiv.r_top.resistance.alias_is(r_any)
    print(f"Set r_top to {r_any}")

    # Solve for r_bottom
    solver.update_superset_cache(rdiv)
    r_bottom = solver.inspect_get_known_supersets(rdiv.r_bottom.resistance)
    assert isinstance(r_bottom, Quantity_Interval_Disjoint)
    print(f"r_bottom: {r_bottom}")
    expected_r_bottom_1 = (v_in / total_current) - r_any
    expected_r_bottom_2 = v_out / total_current
    expected_r_bottom_3 = v_out * r_any / (v_in - v_out)
    print(f"Expected r_bottom subset by voltage: {expected_r_bottom_1}")
    print(f"Expected r_bottom subset by current: {expected_r_bottom_2}")
    print(f"Expected r_bottom subset by voltage and current: {expected_r_bottom_3}")
    assert r_bottom.is_subset_of(expected_r_bottom_1)
    assert r_bottom.is_subset_of(expected_r_bottom_2)
    assert r_bottom.is_subset_of(expected_r_bottom_3)
    # print results
    res_total_current = v_in / (r_any + r_bottom)
    # res_v_out = v_in * r_bottom / (r_any + r_bottom)
    res_v_out = v_in / (1 + r_any / r_bottom)
    solver_total_current = solver.inspect_get_known_supersets(rdiv.max_current)
    solver_v_out = solver.inspect_get_known_supersets(rdiv.v_out)
    print(f"Resulting current {res_total_current} ss! {total_current}")
    print(f"Solver thinks current is {solver_total_current}")
    print(f"Resulting v_out {res_v_out} ss! {v_out}")
    print(f"Solver thinks v_out is {solver_v_out}")

    # check valid result
    assert res_total_current.is_subset_of(total_current)
    if not res_v_out.is_subset_of(v_out) and res_v_out.is_subset_of(
        v_out * L.Range.from_center_rel(1, 0.05)
    ):
        pytest.xfail("Slightly inaccurate, need more symbolic correlation")

    assert res_v_out.is_subset_of(v_out)

    # check solver knowing result
    assert solver_v_out == res_v_out
    assert solver_total_current == res_total_current



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import importlib.util
import logging
import re
from pathlib import Path

import typer

from faebryk.libs.logging import FLOG_FMT, setup_basic_logging

logger = logging.getLogger(__name__)

FLOG_FMT.set(True)


def main(
    filepath: Path = typer.Argument(Path("test")),
    test_name: str = typer.Option(".*", "-k", help="Test name pattern (regex)"),
):
    if not filepath.exists():
        raise ValueError(f"Filepath {filepath} does not exist")
    if not filepath.is_dir():
        filepaths = [filepath]
    else:
        assert filepath.is_dir()
        filepaths = list(filepath.rglob("test_*.py"))

    matches = []
    for fp in filepaths:
        spec = importlib.util.spec_from_file_location("test_module", fp)
        if spec is None:
            continue
        module = importlib.util.module_from_spec(spec)
        if spec.loader is None:
            continue
        try:
            spec.loader.exec_module(module)
        except Exception:
            continue
        for v in vars(module).values():
            if not hasattr(v, "__name__"):
                continue
            if not re.match(test_name, v.__name__):
                continue
            matches.append(v)

    if not matches:
        raise ValueError(f"Test function '{test_name}' not found in {filepaths}")

    setup_basic_logging()

    for test_func in matches:
        if len(matches) > 1:
            logger.info(f"Running {test_func.__name__}")
        test_func()


if __name__ == "__main__":
    typer.run(main)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

from typing import TypeGuard

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.core.moduleinterface import ModuleInterface
from faebryk.libs.library import L
from faebryk.libs.util import once


class _TSwitch(Module):
    T = ModuleInterface

    def __init__(self, t: type[T]):
        super().__init__()
        self.t = t

    @staticmethod
    def is_instance(obj: Module, t: type[T]) -> bool:
        return isinstance(obj, _TSwitch) and issubclass(obj.t, t)


@once  # This means we can use a normal "isinstance" to test for them
def Switch[T: ModuleInterface](interface_type: type[T]):
    class _Switch(_TSwitch):
        def __init__(self) -> None:
            super().__init__(interface_type)

        designator_prefix = L.f_field(F.has_designator_prefix)(
            F.has_designator_prefix.Prefix.S
        )
        attach_to_footprint: F.can_attach_to_footprint_symmetrically

        unnamed = L.list_field(2, interface_type)

        @L.rt_field
        def can_bridge(self):
            return F.can_bridge_defined(*self.unnamed)

        @staticmethod
        def is_instance(obj: Module) -> "TypeGuard[_Switch]":
            return _TSwitch.is_instance(obj, interface_type)

    return _Switch



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import faebryk.library._F as F
from faebryk.core.moduleinterface import ModuleInterface


class can_bridge_defined(F.can_bridge.impl()):
    def __init__(self, in_if: ModuleInterface, out_if: ModuleInterface) -> None:
        super().__init__()
        self.get_in = lambda: in_if
        self.get_out = lambda: out_if



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
from typing import TYPE_CHECKING, Self

from faebryk.core.module import Module
from faebryk.libs.util import not_none

if TYPE_CHECKING:
    from faebryk.libs.picker.picker import PickedPart

logger = logging.getLogger(__name__)


class has_part_picked(Module.TraitT.decless()):
    def __init__(self, part: "PickedPart | None") -> None:
        """
        None = remove part from netlist
        """
        super().__init__()
        if type(self) is has_part_picked and part is None:
            raise ValueError("Use has_part_picked.remove()")
        self.part = part

    @classmethod
    def by_supplier(
        cls, supplier_id: str, supplier_partno: str, manufacturer: str, partno: str
    ) -> Self:
        from faebryk.libs.picker.lcsc import PickedPartLCSC

        match supplier_id:
            case "lcsc":
                return cls(
                    PickedPartLCSC(
                        manufacturer=manufacturer,
                        partno=partno,
                        supplier_partno=supplier_partno,
                    )
                )
            case _:
                raise ValueError(f"Unknown supplier: {supplier_id}")

    def get_part(self) -> "PickedPart":
        return not_none(self.part)

    def try_get_part(self) -> "PickedPart | None":
        return self.part

    @property
    def removed(self) -> bool:
        return self.part is None



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import faebryk.library._F as F
from faebryk.core.core import Namespace


class ElectricLogicGates(Namespace):
    class OR(F.ElectricLogicGate):
        def __init__(self, input_cnt: int):
            super().__init__(input_cnt, 1, F.LogicGate.can_logic_or_gate())

    class NOR(F.ElectricLogicGate):
        def __init__(self, input_cnt: int):
            super().__init__(input_cnt, 1, F.LogicGate.can_logic_nor_gate())

    class NAND(F.ElectricLogicGate):
        def __init__(self, input_cnt: int):
            super().__init__(input_cnt, 1, F.LogicGate.can_logic_nand_gate())

    class XOR(F.ElectricLogicGate):
        def __init__(self, input_cnt: int):
            super().__init__(input_cnt, 1, F.LogicGate.can_logic_xor_gate())



import logging
from pathlib import Path

from atopile.mcp.util import MCPTools
from faebryk.libs.util import root_by_file

logger = logging.getLogger(__name__)

project_tools = MCPTools()


@project_tools.register()
def find_project_from_filepath(absolute_path_to_file: Path) -> Path:
    """
    Find the project root from an .ato file inside that project.
    """

    path = absolute_path_to_file
    logger.info(f"Finding project from filepath: {path}")
    if not path.is_absolute():
        raise ValueError("Path is not absolute")
    if not path.exists():
        raise ValueError(f"Path {path} does not exist")
    if not path.is_file():
        raise ValueError(f"Path {path} is not a file")

    if path.suffix != ".ato":
        raise ValueError("Path is not an ato file")

    return root_by_file("ato.yaml", path.parent)



import re
import sys

MODULE_REGEX = r"^[_a-zA-Z][_a-zA-Z0-9]+$"

module_name = "{{ cookiecutter.project_slug}}"

if not re.match(MODULE_REGEX, module_name):
    print(
        "ERROR: The project slug (%s) is not a valid Python module name. "
        "Please do not use a - and use _ instead" % module_name
    )
    # Exit to cancel project
    sys.exit(1)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT


import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.libs.library import L
from faebryk.libs.units import P


class Inductor(Module):
    unnamed = L.list_field(2, F.Electrical)

    inductance = L.p_field(
        units=P.H,
        likely_constrained=True,
        soft_set=L.Range(100 * P.nH, 1 * P.H),
        tolerance_guess=10 * P.percent,
    )
    max_current = L.p_field(
        units=P.A,
        likely_constrained=True,
        soft_set=L.Range(1 * P.mA, 100 * P.A),
    )
    dc_resistance = L.p_field(
        units=P.Ω,
        soft_set=L.Range(10 * P.mΩ, 100 * P.Ω),
        tolerance_guess=10 * P.percent,
    )
    saturation_current = L.p_field(units=P.A)
    self_resonant_frequency = L.p_field(
        units=P.Hz,
        likely_constrained=True,
        soft_set=L.Range(100 * P.kHz, 1 * P.GHz),
        tolerance_guess=10 * P.percent,
    )

    @L.rt_field
    def pickable(self) -> F.is_pickable_by_type:
        return F.is_pickable_by_type(
            F.is_pickable_by_type.Type.Inductor,
            {
                "inductance": self.inductance,
                "max_current": self.max_current,
                "dc_resistance": self.dc_resistance,
                "saturation_current": self.saturation_current,
                "self_resonant_frequency": self.self_resonant_frequency,
            },
        )

    @L.rt_field
    def can_bridge(self):
        return F.can_bridge_defined(*self.unnamed)

    attach_to_footprint: F.can_attach_to_footprint_symmetrically

    @L.rt_field
    def simple_value_representation(self):
        S = F.has_simple_value_representation_based_on_params_chain.Spec
        return F.has_simple_value_representation_based_on_params_chain(
            S(self.inductance, tolerance=True),
            S(self.self_resonant_frequency),
            S(self.max_current),
            S(self.dc_resistance),
        )

    designator_prefix = L.f_field(F.has_designator_prefix)(
        F.has_designator_prefix.Prefix.L
    )

    # TODO: remove @https://github.com/atopile/atopile/issues/727
    @property
    def p1(self) -> F.Electrical:
        """Signal to one side of the inductor."""
        return self.unnamed[0]

    @property
    def p2(self) -> F.Electrical:
        """Signal to the other side of the inductor."""
        return self.unnamed[1]



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging

import faebryk.library._F as F

logger = logging.getLogger(__name__)


class has_part_removed(F.has_part_picked):
    def __init__(self):
        super().__init__(None)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
from itertools import pairwise
from textwrap import indent
from typing import Callable

import pytest

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.core.solver.algorithm import get_algorithms
from faebryk.core.solver.defaultsolver import DefaultSolver
from faebryk.core.solver.solver import LOG_PICK_SOLVE
from faebryk.core.solver.utils import S_LOG, set_log_level
from faebryk.libs.library import L
from faebryk.libs.picker.picker import (
    NO_PROGRESS_BAR,
    PickError,
    get_pick_tree,
    pick_topologically,
)
from faebryk.libs.test.times import Times
from faebryk.libs.units import P
from faebryk.libs.util import ConfigFlagInt, indented_container
from test.common.resources.fabll_modules.RP2040 import RP2040
from test.common.resources.fabll_modules.RP2040_ReferenceDesign import (
    RP2040_ReferenceDesign,
)

logger = logging.getLogger(__name__)

GROUPS = ConfigFlagInt("GROUPS", 4)
GROUP_SIZE = ConfigFlagInt("GROUP_SIZE", 4)


@pytest.fixture(autouse=True)
def _setup():
    NO_PROGRESS_BAR.set(True)


class _RP2040_Basic(Module):
    rp2040: RP2040
    ldo: F.LDO
    led: F.LED


@pytest.mark.slow
@pytest.mark.usefixtures("setup_project_config")
@pytest.mark.parametrize(
    "module_type",
    [
        _RP2040_Basic,
        RP2040_ReferenceDesign,
        lambda: F.MultiCapacitor(10),
    ],
)
def test_performance_pick_real_module(module_type: Callable[[], Module]):
    timings = Times()

    app = module_type()
    timings.add("construct")

    F.is_bus_parameter.resolve_bus_parameters(app.get_graph())
    timings.add("resolve bus params")

    pick_tree = get_pick_tree(app)
    timings.add("pick tree")

    solver = DefaultSolver()

    with timings.as_global("pick"):
        pick_topologically(pick_tree, solver)

    logger.info(f"\n{timings}")


@pytest.mark.slow
@pytest.mark.usefixtures("setup_project_config")
def test_performance_pick_rc_formulas():
    _GROUPS = int(GROUPS)
    _GROUP_SIZE = int(GROUP_SIZE)
    INCREASE = 10 * P.percent
    TOLERANCE = 20 * P.percent

    class App(Module):
        alias_res = L.list_field(_GROUPS, F.Resistor)
        res = L.list_field(_GROUPS * _GROUP_SIZE, F.Resistor)

        def __preinit__(self):
            increase = L.Range.from_center_rel(INCREASE, TOLERANCE) + L.Single(
                100 * P.percent
            )

            for i in range(_GROUPS):
                for m1, m2 in pairwise(self.res[i::_GROUPS]):
                    m2.resistance.constrain_subset(m1.resistance * increase)
                    # solver doesn't do equation reordering, so we need to reverse
                    m1.resistance.constrain_subset(m2.resistance / increase)
                self.alias_res[i].resistance.alias_is(self.res[i].resistance)

    timings = Times(multi_sample_strategy=Times.MultiSampleStrategy.ALL)

    app = App()
    timings.add("construct")

    F.is_bus_parameter.resolve_bus_parameters(app.get_graph())
    timings.add("resolve bus params")

    pick_tree = get_pick_tree(app)
    timings.add("pick tree")

    solver = DefaultSolver()
    try:
        with timings.as_global("pick"):
            pick_topologically(pick_tree, solver)
    except PickError as e:
        logger.error(f"Error picking: {e.args[0]}")
        params = {
            m.get_full_name(): "\n" + indent(m.pretty_params(solver), prefix="  ")
            for m in app.get_children_modules(direct_only=True, types=Module)
        }
        params = {k: v for k, v in sorted(params.items(), key=lambda t: t[0])}
        logger.info(f"Params:{indented_container(params, use_repr=False)}")
        S_LOG.set(True, force=True)
        LOG_PICK_SOLVE.set(True, force=True)
        set_log_level(logging.DEBUG)
        solver.update_superset_cache(app)
        # assert False
        return
    finally:

        def _is_algo(
            k: str, dirty: bool | None = None, terminal: bool | None = None
        ) -> bool:
            if "run_iteration:" not in k:
                return False
            if ":setup" in k or ":close" in k:
                return False
            if "clean" not in k and "dirty" not in k:
                return False
            if dirty is not None:
                if dirty and "dirty" not in k:
                    return False
                if not dirty and "clean" not in k:
                    return False
            if terminal is not None:
                if terminal and " terminal" not in k:
                    return False
                if not terminal and "non-terminal" not in k:
                    return False
            return True

        def _make_algo_group(dirty: bool | None = None, terminal: bool | None = None):
            dirty_str = "" if dirty is None else "dirty " if dirty else "clean "
            terminal_str = (
                "" if terminal is None else "terminal " if terminal else "non-terminal "
            )
            timings.make_group(
                f"{dirty_str}{terminal_str}algos",
                lambda k: _is_algo(k, dirty=dirty, terminal=terminal),
            )

        timings.add_seperator()
        for algo in get_algorithms():
            timings.make_group("Total " + algo.name, lambda k: algo.name + " " in k)
        timings.add_seperator()
        for i in [None, True, False]:
            for j in [None, True, False]:
                _make_algo_group(dirty=i, terminal=j)
        timings.add_seperator()
        timings.make_group(
            "mutator setup",
            lambda k: "run_iteration:setup" in k or "run_iteration:close" in k,
        )
        timings.make_group("backend wait", lambda k: "fetch parts" in k)
        timings.make_group("solver", lambda k: "algos" == k)
        logger.info(f"\n{timings.to_str(force_unit='ms')}")

    picked_values = {
        m.get_full_name(): str(m.resistance.try_get_literal()) for m in app.res
    }
    logger.info(f"Picked values: {indented_container(picked_values)}")

    pick_time = timings.get_formatted("pick", strat=Times.MultiSampleStrategy.ACC)
    logger.info(f"Pick duration {_GROUPS}x{_GROUP_SIZE}: {pick_time}")



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import faebryk.library._F as F


class has_pcb_position_defined(F.has_pcb_position.impl()):
    def __init__(self, position: F.has_pcb_position.Point) -> None:
        super().__init__()
        self.position = position

    def get_position(self) -> F.has_pcb_position.Point:
        return self.position



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging

import faebryk.library._F as F  # noqa: F401
from faebryk.libs.library import L  # noqa: F401
from faebryk.libs.units import P  # noqa: F401

logger = logging.getLogger(__name__)


class Winbond_Elec_W25Q128JVSIQ(F.SPIFlash):
    """
    TODO: Docstring describing your module

    128Mbit SOIC-8-208mil
    NOR FLASH ROHS
    """

    # ----------------------------------------
    #     modules, interfaces, parameters
    # ----------------------------------------

    # ----------------------------------------
    #                 traits
    # ----------------------------------------

    explicit_part = L.f_field(F.has_explicit_part.by_supplier)("C97521")

    designator_prefix = L.f_field(F.has_designator_prefix)(
        F.has_designator_prefix.Prefix.U
    )

    @L.rt_field
    def pin_association_heuristic(self):
        return F.has_pin_association_heuristic_lookup_table(
            mapping={
                self.qspi.chip_select.line: ["CS#", "~{CS}"],
                self.qspi.data[0].line: ["DO"],
                self.qspi.data[2].line: ["IO2"],
                self.power.lv: ["GND"],
                self.qspi.data[1].line: ["DI"],
                self.qspi.clock.line: ["CLK"],
                self.qspi.data[3].line: ["IO3"],
                self.power.hv: ["VCC"],
            },
            accept_prefix=True,
            case_sensitive=False,
        )

    @L.rt_field
    def decoupled(self):
        return F.can_be_decoupled_rails(self.power)

    def __preinit__(self):
        # ------------------------------------
        #           connections
        # ------------------------------------

        # ------------------------------------
        #          parametrization
        # ------------------------------------
        pass



import os
import shutil
import sys
from pathlib import Path

import pytest

from faebryk.libs.util import repo_root as _repo_root
from faebryk.libs.util import run_live

# Get the examples directory relative to this test file
EXAMPLES_DIR = _repo_root() / "examples"

FABLL_EXAMPLES = [p for p in EXAMPLES_DIR.glob("*.py") if p.is_file()]
ATO_EXAMPLES = [p for p in EXAMPLES_DIR.glob("*.ato") if p.is_file()]

SLOW = {"ch2_8_mcu"}
XFAILURES = {"ch2_8_mcu": "Pin matching issues"}


def _get_marks(example: Path):
    return [
        *(
            [pytest.mark.xfail(reason=XFAILURES[example.stem])]
            if example.stem in XFAILURES
            else []
        ),
        *([pytest.mark.slow] if example.stem in SLOW else []),
    ]


@pytest.mark.parametrize(
    "example",
    [
        pytest.param(example, marks=_get_marks(example))
        for example in FABLL_EXAMPLES + ATO_EXAMPLES
    ],
    ids=lambda p: p.stem,
)
def test_examples_build(
    example: Path, tmp_path: Path, repo_root: Path, save_tmp_path_on_failure: None
):
    assert example.exists()

    example_copy = tmp_path / example.name
    example_copy.write_text(example.read_text(encoding="utf-8"), encoding="utf-8")
    example = example_copy

    # Copy dependencies to the tmp dir directly because standalone mode doens't include
    example_modules = repo_root / "test" / "common" / "resources" / ".ato" / "modules"
    for item in example_modules.glob("*"):
        if item.is_dir():
            shutil.copytree(item, tmp_path / item.name)
        else:
            shutil.copy(item, tmp_path / item.name)

    _, stderr, _ = run_live(
        [
            sys.executable,
            "-m",
            "atopile",
            "build",
            "--standalone",
            f"{example}:App",
        ],
        env={**os.environ, "NONINTERACTIVE": "1"},
        cwd=tmp_path,
        stdout=print,
        stderr=print,
    )

    # TODO: add a strict mode to the CLI
    assert "Build successful! 🚀" in stderr
    assert stderr.count("✓") >= 1
    assert stderr.count("✗") == 0

    # expected warnings:
    # - missing kicad-cli for '3d-model' target (in CI only)
    assert stderr.count("⚠") in (0, 1)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT
import logging
from typing import TYPE_CHECKING, Callable, Iterable

from faebryk.core.cpp import GraphInterfaceModuleSibling
from faebryk.core.node import Node, NodeException, f_field
from faebryk.core.parameter import Parameter
from faebryk.core.trait import Trait
from faebryk.libs.exceptions import accumulate
from faebryk.libs.util import cast_assert, unique_ref

if TYPE_CHECKING:
    from faebryk.core.moduleinterface import ModuleInterface

logger = logging.getLogger(__name__)


class ModuleException(NodeException):
    def __init__(self, module: "Module", *args: object) -> None:
        self.module = module
        super().__init__(module, *args)


class Module(Node):
    class TraitT(Trait): ...

    specializes = f_field(GraphInterfaceModuleSibling)(is_parent=False)
    specialized = f_field(GraphInterfaceModuleSibling)(is_parent=True)

    def get_most_special(self) -> "Module":
        specialers = {
            specialer
            for specialer_gif in self.specialized.get_gif_edges()
            if (specialer := specialer_gif.node) is not self
            and isinstance(specialer, Module)
        }
        if not specialers:
            return self

        specialest_next = unique_ref(
            specialer.get_most_special() for specialer in specialers
        )

        assert len(specialest_next) == 1, (
            f"Ambiguous specialest {specialest_next} for {self}"
        )
        return next(iter(specialest_next))

    def get_children_modules[T: Module](
        self: Node,
        types: type[T] | tuple[type[T], ...],
        most_special: bool = True,
        direct_only: bool = False,
        include_root: bool = False,
        f_filter: Callable[[T], bool] | None = None,
        sort: bool = True,
    ) -> set[T]:
        out = self.get_children(
            direct_only=direct_only,
            types=types,
            include_root=include_root,
            f_filter=f_filter,
            sort=sort,
        )
        out_specialized = {
            n.get_most_special()
            for n in self.get_children(
                direct_only=direct_only,
                types=Module,
                include_root=include_root,
                f_filter=lambda x: x.get_most_special() != x,
                sort=sort,
            )
        }
        if most_special:
            special_out = set()
            todo = out_specialized
            # TODO can be done more efficiently by just allowing in the graph search
            # specialize edges
            for n in out:
                # Filter out children of specialized modules
                has_specialized_parent = n.get_parent_f(
                    lambda x: isinstance(x, Module) and x.get_most_special() != x
                )
                if has_specialized_parent:
                    continue
                n_special = n.get_most_special()
                # Non-special can just pass
                if n_special is n:
                    special_out.add(n_special)
                    continue
                # Already processed
                if n_special in out | special_out:
                    continue
                # To process children
                todo.add(n)

            for n_special in todo:
                special_out.update(
                    n_special.get_children_modules(
                        types=types,
                        most_special=True,
                        direct_only=direct_only,
                        include_root=True,
                        f_filter=f_filter,
                        sort=sort,
                    )
                )
            out = special_out

        return out

    class InvalidSpecializationError(Exception):
        """Cannot specialize module with special"""

        def __init__(
            self, message: str, *args, module: "Module", special: "Module", **kwargs
        ):
            self.message = message
            self.module = module
            self.special = special
            super().__init__(message, *args, **kwargs)

    def specialize[T: Module](
        self,
        special: T,
        matrix: list[tuple["ModuleInterface", "ModuleInterface"]] | None = None,
        attach_to: Node | None = None,
    ) -> T:
        from faebryk.core.moduleinterface import ModuleInterface
        from faebryk.core.parameter import Parameter

        logger.debug(f"Specializing Module {self} with {special}" + " " + "=" * 20)

        def get_node_prop_matrix[N: Node](sub_type: type[N]):
            return list(self.zip_children_by_name_with(special, sub_type).values())

        if matrix is None:
            matrix = get_node_prop_matrix(ModuleInterface)

        # TODO add warning if not all src interfaces used
        param_matrix = get_node_prop_matrix(Parameter)

        err_acc = accumulate(self.InvalidSpecializationError)

        for src, dst in matrix:
            with err_acc.collect():
                if src is None:
                    continue
                if dst is None:
                    raise self.InvalidSpecializationError(
                        f"Special module misses interface: {src.get_name()}",
                        module=self,
                        special=special,
                    )
                src.specialize(dst)

        for src, dst in param_matrix:
            with err_acc.collect():
                if src is None:
                    continue
                if dst is None:
                    raise self.InvalidSpecializationError(
                        f"Special module misses parameter: {src.get_name()}",
                        module=self,
                        special=special,
                    )
                dst.alias_is(src)

        err_acc.raise_errors()

        # TODO this cant work
        # for t in self.traits:
        #    # TODO needed?
        #    if special.has_trait(t.trait):
        #        continue
        #    special.add(t)

        self.specialized.connect(special.specializes)

        # Attach to new parent
        has_parent = special.get_parent() is not None
        assert not has_parent or attach_to is None
        if not has_parent:
            if attach_to:
                attach_to.add(special, container=attach_to.specialized_)
            else:
                gen_parent = self.get_parent()
                if gen_parent:
                    cast_assert(Node, gen_parent[0]).add(
                        special, name=f"{gen_parent[1]}_specialized"
                    )

        return special

    @staticmethod
    def connect_all_interfaces_by_name(
        src: Iterable["Module"] | "Module",
        dst: Iterable["Module"] | "Module",
        allow_partial: bool = False,
    ):
        from faebryk.core.moduleinterface import ModuleInterface

        if isinstance(src, Module):
            src = [src]
        if isinstance(dst, Module):
            dst = [dst]

        for src_, dst_ in zip(src, dst):
            for k, (src_m, dst_m) in src_.zip_children_by_name_with(
                dst_, ModuleInterface
            ).items():
                # TODO: careful this also connects runtime children
                # for now skip stuff prefixed with _
                if k.startswith("_"):
                    continue
                if src_m is None or dst_m is None:
                    if not allow_partial:
                        raise Exception(f"Node with name {k} not present in both")
                    continue
                src_m.connect(dst_m)

    def connect_interfaces_by_name(self, *dst: "Module", allow_partial: bool = False):
        type(self).connect_all_interfaces_by_name(
            self, dst, allow_partial=allow_partial
        )

    def get_parameters(self) -> list[Parameter]:
        return list(self.get_children(types=Parameter, direct_only=True))

    # TODO get rid of this abomination
    @property
    def reference_shim(self):
        from faebryk.library.has_single_electric_reference import (
            has_single_electric_reference,
        )

        return self.get_trait(has_single_electric_reference).get_reference()



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

from enum import Enum, auto

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.libs.library import L
from faebryk.libs.units import P


class Filter(Module):
    class Response(Enum):
        LOWPASS = auto()
        HIGHPASS = auto()
        BANDPASS = auto()
        BANDSTOP = auto()
        OTHER = auto()

    cutoff_frequency = L.p_field(
        units=P.Hz,
        likely_constrained=True,
        domain=L.Domains.Numbers.REAL(),
        soft_set=L.Range(0 * P.Hz, 1000 * P.Hz),
    )
    order = L.p_field(
        domain=L.Domains.Numbers.NATURAL(),
        soft_set=L.Range(2, 10),
        guess=2,
    )
    response = L.p_field(
        domain=L.Domains.ENUM(Response),
    )

    in_: F.Signal
    out: F.Signal



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
from typing import TYPE_CHECKING

import faebryk.library._F as F  # noqa: F401
from faebryk.libs.library import L  # noqa: F401
from faebryk.libs.units import P  # noqa: F401

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from test.common.resources.fabll_modules.RP2040 import RP2040


class RP2040Pinmux(F.Pinmux):
    """
    Raspberry PI RP2040 Pinmux
    """

    # ----------------------------------------
    #     modules, interfaces, parameters
    # ----------------------------------------

    # ----------------------------------------
    #                 traits
    # ----------------------------------------
    def __init__(self, mcu: "RP2040"):
        super().__init__()
        self._mcu = mcu

    def __preinit__(self):
        # ------------------------------------
        #          parametrization
        # ------------------------------------

        # ------------------------------------
        #           connections
        # ------------------------------------
        pass

    # RP2040 Pin Multiplexing Functions
    # +-------------+----------------------------------------------------------+
    # | Function    | Description                                              |
    # +-------------+----------------------------------------------------------+
    # | SPIx        | Connect internal PL022 SPI peripherals to GPIO           |
    # | UARTx       | Connect internal PL011 UART peripherals to GPIO          |
    # | I2Cx        | Connect internal DW I2C peripherals to GPIO              |
    # | PWMx A/B    | Connect PWM slices to GPIO (8 slices, 2 outputs each)    |
    # |             | B pin can be used as input for frequency/duty cycle      |
    # | SIO (F5)    | Software control of GPIO via single-cycle IO block       |
    # |             | Always connected for input, must be selected for output  |
    # | PIOx (F6,F7)| Connect programmable IO blocks to GPIO                   |
    # |             | Flexible interface placement, always connected for input |
    # | CLOCK GPINx | General purpose clock inputs                             |
    # |             | Can be routed to internal clock domains or counters      |
    # | CLOCK GPOUTx| General purpose clock outputs                            |
    # |             | Can output internal clocks to GPIOs with optional divide |
    # | USB Control | OVCUR DET, VBUS DET, VBUS EN                             |
    # |             | USB power control signals to/from internal controller    |
    # +-------------+----------------------------------------------------------+

    # TODO SIO (F5):

    def _get_ios(self) -> list[F.Electrical]:
        return self._mcu.io

    def _get_matrix(self) -> dict[F.Electrical, list[F.Electrical | None]]:
        x = self._mcu
        return {
            x.io[0]: [
                x.spi[0].miso.line,
                x.uart[0].base_uart.tx.line,
                x.i2c[0].sda.line,
                x.pwm[0].A.line,
                x.gpio[0].line,
                x.pio[0].line,
                x.pio[1].line,
                None,
                x.usb_power_control.ovcur_det.line,
            ],
            x.io[1]: [
                x.spi[0].cs.line,
                x.uart[0].base_uart.rx.line,
                x.i2c[0].scl.line,
                x.pwm[0].B.line,
                x.gpio[1].line,
                x.pio[0].line,
                x.pio[1].line,
                None,
                x.usb_power_control.vbus_det.line,
            ],
            x.io[2]: [
                x.spi[0].sclk.line,
                x.uart[0].cts.line,
                x.i2c[1].sda.line,
                x.pwm[1].A.line,
                x.gpio[2].line,
                x.pio[0].line,
                x.pio[1].line,
                None,
                x.usb_power_control.vbus_en.line,
            ],
            x.io[3]: [
                x.spi[0].mosi.line,
                x.uart[0].rts.line,
                x.i2c[1].scl.line,
                x.pwm[1].B.line,
                x.gpio[3].line,
                x.pio[0].line,
                x.pio[1].line,
                None,
                x.usb_power_control.ovcur_det.line,
            ],
            x.io[4]: [
                x.spi[0].miso.line,
                x.uart[1].base_uart.tx.line,
                x.i2c[0].sda.line,
                x.pwm[2].A.line,
                x.gpio[4].line,
                x.pio[0].line,
                x.pio[1].line,
                None,
                x.usb_power_control.vbus_det.line,
            ],
            x.io[5]: [
                x.spi[0].cs.line,
                x.uart[1].base_uart.rx.line,
                x.i2c[0].scl.line,
                x.pwm[2].B.line,
                x.gpio[5].line,
                x.pio[0].line,
                x.pio[1].line,
                None,
                x.usb_power_control.vbus_en.line,
            ],
            x.io[6]: [
                x.spi[0].sclk.line,
                x.uart[1].cts.line,
                x.i2c[1].sda.line,
                x.pwm[3].A.line,
                x.gpio[6].line,
                x.pio[0].line,
                x.pio[1].line,
                None,
                x.usb_power_control.ovcur_det.line,
            ],
            x.io[7]: [
                x.spi[0].mosi.line,
                x.uart[1].rts.line,
                x.i2c[1].scl.line,
                x.pwm[3].B.line,
                x.gpio[7].line,
                x.pio[0].line,
                x.pio[1].line,
                None,
                x.usb_power_control.vbus_det.line,
            ],
            x.io[8]: [
                x.spi[1].miso.line,
                x.uart[0].base_uart.tx.line,
                x.i2c[0].sda.line,
                x.pwm[4].A.line,
                x.gpio[8].line,
                x.pio[0].line,
                x.pio[1].line,
                None,
                x.usb_power_control.vbus_en.line,
            ],
            x.io[9]: [
                x.spi[1].cs.line,
                x.uart[0].base_uart.rx.line,
                x.i2c[0].scl.line,
                x.pwm[4].B.line,
                x.gpio[9].line,
                x.pio[0].line,
                x.pio[1].line,
                None,
                x.usb_power_control.ovcur_det.line,
            ],
            x.io[10]: [
                x.spi[1].sclk.line,
                x.uart[0].cts.line,
                x.i2c[1].sda.line,
                x.pwm[5].A.line,
                x.gpio[10].line,
                x.pio[0].line,
                x.pio[1].line,
                None,
                x.usb_power_control.vbus_det.line,
            ],
            x.io[11]: [
                x.spi[1].mosi.line,
                x.uart[0].rts.line,
                x.i2c[1].scl.line,
                x.pwm[5].B.line,
                x.gpio[11].line,
                x.pio[0].line,
                x.pio[1].line,
                None,
                x.usb_power_control.vbus_en.line,
            ],
            x.io[12]: [
                x.spi[1].miso.line,
                x.uart[1].base_uart.tx.line,
                x.i2c[0].sda.line,
                x.pwm[6].A.line,
                x.gpio[12].line,
                x.pio[0].line,
                x.pio[1].line,
                None,
                x.usb_power_control.ovcur_det.line,
            ],
            x.io[13]: [
                x.spi[1].cs.line,
                x.uart[1].base_uart.rx.line,
                x.i2c[0].scl.line,
                x.pwm[6].B.line,
                x.gpio[13].line,
                x.pio[0].line,
                x.pio[1].line,
                None,
                x.usb_power_control.vbus_det.line,
            ],
            x.io[14]: [
                x.spi[1].sclk.line,
                x.uart[1].cts.line,
                x.i2c[1].sda.line,
                x.pwm[7].A.line,
                x.gpio[14].line,
                x.pio[0].line,
                x.pio[1].line,
                None,
                x.usb_power_control.vbus_en.line,
            ],
            x.io[15]: [
                x.spi[1].mosi.line,
                x.uart[1].rts.line,
                x.i2c[1].scl.line,
                x.pwm[7].B.line,
                x.gpio[15].line,
                x.pio[0].line,
                x.pio[1].line,
                None,
                x.usb_power_control.ovcur_det.line,
            ],
            x.io[16]: [
                x.spi[0].miso.line,
                x.uart[0].base_uart.tx.line,
                x.i2c[0].sda.line,
                x.pwm[0].A.line,
                x.gpio[16].line,
                x.pio[0].line,
                x.pio[1].line,
                None,
                x.usb_power_control.vbus_det.line,
            ],
            x.io[17]: [
                x.spi[0].cs.line,
                x.uart[0].base_uart.rx.line,
                x.i2c[0].scl.line,
                x.pwm[0].B.line,
                x.gpio[17].line,
                x.pio[0].line,
                x.pio[1].line,
                None,
                x.usb_power_control.vbus_en.line,
            ],
            x.io[18]: [
                x.spi[0].sclk.line,
                x.uart[0].cts.line,
                x.i2c[1].sda.line,
                x.pwm[1].A.line,
                x.gpio[18].line,
                x.pio[0].line,
                x.pio[1].line,
                None,
                x.usb_power_control.ovcur_det.line,
            ],
            x.io[19]: [
                x.spi[0].mosi.line,
                x.uart[0].rts.line,
                x.i2c[1].scl.line,
                x.pwm[1].B.line,
                x.gpio[19].line,
                x.pio[0].line,
                x.pio[1].line,
                None,
                x.usb_power_control.vbus_det.line,
            ],
            x.io[20]: [
                x.spi[0].miso.line,
                x.uart[1].base_uart.tx.line,
                x.i2c[0].sda.line,
                x.pwm[2].A.line,
                x.gpio[20].line,
                x.pio[0].line,
                x.pio[1].line,
                x.clock_in[0].line,
                x.usb_power_control.vbus_en.line,
            ],
            x.io[21]: [
                x.spi[0].cs.line,
                x.uart[1].base_uart.rx.line,
                x.i2c[0].scl.line,
                x.pwm[2].B.line,
                x.gpio[21].line,
                x.pio[0].line,
                x.pio[1].line,
                x.clock_out[0].line,
                x.usb_power_control.ovcur_det.line,
            ],
            x.io[22]: [
                x.spi[0].sclk.line,
                x.uart[1].cts.line,
                x.i2c[1].sda.line,
                x.pwm[3].A.line,
                x.gpio[22].line,
                x.pio[0].line,
                x.pio[1].line,
                x.clock_in[1].line,
                x.usb_power_control.vbus_det.line,
            ],
            x.io[23]: [
                x.spi[0].mosi.line,
                x.uart[1].rts.line,
                x.i2c[1].scl.line,
                x.pwm[3].B.line,
                x.gpio[23].line,
                x.pio[0].line,
                x.pio[1].line,
                x.clock_out[1].line,
                x.usb_power_control.vbus_en.line,
            ],
            x.io[24]: [
                x.spi[1].miso.line,
                x.uart[0].base_uart.tx.line,
                x.i2c[0].sda.line,
                x.pwm[4].A.line,
                x.gpio[24].line,
                x.pio[0].line,
                x.pio[1].line,
                x.clock_out[2].line,
                x.usb_power_control.ovcur_det.line,
            ],
            x.io[25]: [
                x.spi[1].cs.line,
                x.uart[0].base_uart.rx.line,
                x.i2c[0].scl.line,
                x.pwm[4].B.line,
                x.gpio[25].line,
                x.pio[0].line,
                x.pio[1].line,
                x.clock_out[3].line,
                x.usb_power_control.vbus_det.line,
            ],
            x.io[26]: [
                x.spi[1].sclk.line,
                x.uart[0].cts.line,
                x.i2c[1].sda.line,
                x.pwm[5].A.line,
                x.gpio[26].line,
                x.pio[0].line,
                x.pio[1].line,
                None,
                x.usb_power_control.vbus_en.line,
            ],
            x.io[27]: [
                x.spi[1].mosi.line,
                x.uart[0].rts.line,
                x.i2c[1].scl.line,
                x.pwm[5].B.line,
                x.gpio[27].line,
                x.pio[0].line,
                x.pio[1].line,
                None,
                x.usb_power_control.ovcur_det.line,
            ],
            x.io[28]: [
                x.spi[1].miso.line,
                x.uart[1].base_uart.tx.line,
                x.i2c[0].sda.line,
                x.pwm[6].A.line,
                x.gpio[28].line,
                x.pio[0].line,
                x.pio[1].line,
                None,
                x.usb_power_control.vbus_det.line,
            ],
            x.io[29]: [
                x.spi[1].cs.line,
                x.uart[1].base_uart.rx.line,
                x.i2c[0].scl.line,
                x.pwm[6].B.line,
                x.gpio[29].line,
                x.pio[0].line,
                x.pio[1].line,
                None,
                x.usb_power_control.vbus_en.line,
            ],
        }



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
import re
from pathlib import Path

from faebryk.core.trait import Trait

logger = logging.getLogger(__name__)

s = r"\s*"


def p_group(pattern: str, name: str | None = None):
    if not name:
        return pattern
    return rf"(?P<{name}>{pattern})"


def p_string(name: str | None = None):
    inner = p_group(r'([^"]*?)', name)
    return rf'"{inner}"'


def p_assignment(key_name: str | None = None, value_name: str | None = None):
    return rf"{p_group(r'\w+', key_name)}{s}={s}{p_string(value_name)}{s}"


class AtoCodeParse:
    """
    Ghetto ato code parser. Only good for quick prototyping.
    Better use Bob.
    """

    class TraitNotFound(Exception):
        pass

    class ComponentFile:
        def __init__(self, content: str | Path):
            if isinstance(content, Path):
                self.ato = content.read_text("utf-8")
            else:
                self.ato = content

            self.ato_no_comments = re.sub(r"#.*", "", self.ato)

        def parse_trait(
            self,
            trait: str,
        ) -> tuple[str | None, dict[str, str]]:
            trait_match = re.search(
                rf"trait {trait}(?P<constructor>::[a-zA-Z0-9_]+)?(?P<args><.+?>)?",
                self.ato_no_comments,
                re.DOTALL,
            )
            if not trait_match:
                raise AtoCodeParse.TraitNotFound(f"Could not find {trait} trait")
            args_stmt = trait_match.group("args") or ""
            m_args = re.finditer(
                f"{p_assignment(key_name='k', value_name='v')}[,>]", args_stmt
            )
            args = {match.group("k"): match.group("v") for match in m_args if match}

            parsed_constructor = (
                trait_match.group("constructor").removeprefix("::")
                if trait_match.group("constructor")
                else None
            )

            return parsed_constructor, args

        def parse_trait_args(
            self,
            trait: str,
            require_args: list[str],
            require_constructor: str | None = None,
        ) -> dict[str, str]:
            trait_constructor, trait_args = self.parse_trait(trait)

            missing = [arg for arg in require_args if arg not in trait_args]
            if missing:
                raise AtoCodeParse.TraitNotFound(
                    f"Missing required arguments for trait {trait}: {missing}"
                )

            if require_constructor and trait_constructor != require_constructor:
                raise AtoCodeParse.TraitNotFound(
                    f"Constructor mismatch for trait"
                    f" {trait}: {trait_constructor} != {require_constructor}"
                )

            return trait_args

        def get_trait[T: Trait](self, trait: type[T]) -> T:
            constructor, args = self.parse_trait(trait.__name__)
            f = trait if constructor is None else getattr(trait, constructor)
            return f(**args)

        def try_get_trait[T: Trait](self, trait: type[T]) -> T | None:
            try:
                return self.get_trait(trait)
            except AtoCodeParse.TraitNotFound:
                return None

        def parse_docstring(self) -> str:
            docstring = re.search(
                r'component [a-zA-Z0-9_]+:\n    """(.*?)"""',
                self.ato_no_comments,
                re.DOTALL,
            )
            if docstring:
                return docstring.group(1)
            else:
                return ""



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT


import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.libs.library import L
from faebryk.libs.units import P


class Crystal_Oscillator(Module):
    # ----------------------------------------
    #     modules, interfaces, parameters
    # ----------------------------------------
    crystal: F.Crystal
    capacitors = L.list_field(2, F.Capacitor)
    current_limiting_resistor: F.Resistor

    xtal_if: F.XtalIF

    # ----------------------------------------
    #               parameters
    # ----------------------------------------
    # https://blog.adafruit.com/2012/01/24/choosing-the-right-crystal-and-caps-for-your-design/
    # http://www.st.com/internet/com/TECHNICAL_RESOURCES/TECHNICAL_LITERATURE/APPLICATION_NOTE/CD00221665.pdf
    _STRAY_CAPACITANCE = L.Range(1 * P.pF, 5 * P.pF)

    @L.rt_field
    def capacitance(self):
        return (self.crystal.load_capacitance - self._STRAY_CAPACITANCE) * 2

    def __preinit__(self):
        for cap in self.capacitors:
            cap.capacitance.constrain_subset(self.capacitance)

        self.current_limiting_resistor.allow_removal_if_zero()
        self.crystal.gnd.connect(self.xtal_if.gnd)
        self.crystal.unnamed[0].connect_via(self.capacitors[0], self.xtal_if.gnd)
        self.crystal.unnamed[1].connect_via(self.capacitors[1], self.xtal_if.gnd)

        self.crystal.unnamed[0].connect_via(
            self.current_limiting_resistor, self.xtal_if.xout
        )
        self.crystal.unnamed[1].connect(self.xtal_if.xin)

    @L.rt_field
    def can_bridge(self):
        return F.can_bridge_defined(self.xtal_if.xin, self.xtal_if.xout)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import faebryk.library._F as F


class has_kicad_footprint_equal_ifs(F.has_kicad_footprint.impl()):
    def get_pin_names(self):
        return self.obj.get_trait(F.has_equal_pins).get_pin_map()



import logging

import typer

from atopile.telemetry import capture

logger = logging.getLogger(__name__)


lsp_app = typer.Typer(rich_markup_mode="rich")


@lsp_app.command()
@capture("cli:lsp_start", "cli:lsp_end")
def start():
    from atopile.lsp import LSP_SERVER

    LSP_SERVER.start_io()



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
from enum import Enum, auto
from typing import TYPE_CHECKING, Sequence

import faebryk.library._F as F
from faebryk.core.moduleinterface import ModuleInterface
from faebryk.libs.geometry.basic import Geometry
from faebryk.libs.util import KeyErrorAmbiguous, KeyErrorNotFound, find

if TYPE_CHECKING:
    from faebryk.exporters.pcb.kicad.transformer import PCB_Transformer

logger = logging.getLogger(__name__)


class has_pcb_routing_strategy_greedy_direct_line(F.has_pcb_routing_strategy.impl()):
    class Topology(Enum):
        STAR = auto()
        DIRECT = auto()
        # CHAIN = auto()

    def __init__(
        self,
        topology: Topology = Topology.DIRECT,
        extra_mifs: Sequence[ModuleInterface] | None = None,
        extra_pads: Sequence[F.Pad] | None = None,
    ) -> None:
        super().__init__()
        self.topology = topology
        self.extra_mifs = extra_mifs or []
        self.extra_pads = extra_pads or []

    def calculate(self, transformer: "PCB_Transformer"):
        from faebryk.exporters.pcb.routing.util import (
            DEFAULT_TRACE_WIDTH,
            Path,
            Route,
            get_internal_nets_of_node,
            get_pads_pos_of_mifs,
            group_pads_that_are_connected_already,
        )

        node = self.obj
        nets = get_internal_nets_of_node(node)

        logger.debug(f"Routing {node} {'-' * 40}")
        # TODO avoid crossing pads
        # might make this very complex though

        def get_route_for_mifs_in_net(mifs) -> Route | None:
            pads = get_pads_pos_of_mifs(mifs + self.extra_mifs, self.extra_pads)

            unique_layers = {layer for pos in pads.values() for layer in pos[3]}
            try:
                layer = find(
                    unique_layers,
                    lambda layer: all(layer in pos[3] for pos in pads.values()),
                )
            except KeyErrorAmbiguous as e:
                layer = e.duplicates[0]
            except KeyErrorNotFound as e:
                raise NotImplementedError(
                    "Only support pads that share at least one layer"
                ) from e

            grouped_pads = group_pads_that_are_connected_already(pads)

            if len(grouped_pads) < 2:
                return None

            logger.debug(f"Routing pads: {pads}")

            def _get_pad_pos(pad: F.Pad):
                """
                overwrite layer set with the layer we chose to route on
                """
                x, y, r, _ = pads[pad]
                return x, y, r, layer

            def get_route_for_net_star():
                # filter pads that are already connected
                fpads = {
                    (pad := next(iter(pad_group))): _get_pad_pos(pad)
                    for pad_group in grouped_pads
                }

                center = Geometry.average([pos for _, pos in fpads.items()])

                path = Path()
                for _, pos in fpads.items():
                    path.add(Path.Line(DEFAULT_TRACE_WIDTH, layer, pos, center))

                return Route(pads=fpads.keys(), path=path)

            def get_route_for_direct():
                sets = [{_get_pad_pos(pad) for pad in group} for group in grouped_pads]

                path = Path()

                while len(sets) > 1:
                    # find closest pads
                    closest = min(
                        (
                            (set1, set2, Geometry.distance_euclid(p1, p2), [p1, p2])
                            for set1 in sets
                            for set2 in sets
                            for p1 in set1
                            for p2 in set2
                            if set1 != set2
                        ),
                        key=lambda t: t[2],
                    )

                    # merge closest pads
                    sets.remove(closest[0])
                    sets.remove(closest[1])
                    sets.append(closest[0].union(closest[1]))

                    path.add(
                        Path.Track(
                            width=DEFAULT_TRACE_WIDTH, layer=layer, points=closest[3]
                        )
                    )

                return Route(pads=pads.keys(), path=path)

            if self.topology == self.Topology.STAR:
                return get_route_for_net_star()
            elif self.topology == self.Topology.DIRECT:
                return get_route_for_direct()
            else:
                raise NotImplementedError

        return [
            route
            for net, mifs in nets.items()
            if net and (route := get_route_for_mifs_in_net(mifs))
        ]



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

from enum import StrEnum

import faebryk.library._F as F
from faebryk.libs.library import L
from faebryk.libs.smd import SMDSize


class SMDTwoPin(F.Footprint):
    class Type(StrEnum):
        Resistor = "R"
        Capacitor = "C"
        Inductor = "L"

    def __init__(self, size: SMDSize, type: Type) -> None:
        super().__init__()
        self._size = size
        self._type = type

    pins = L.list_field(2, F.Pad)

    class _has_kicad_footprint(F.has_kicad_footprint_equal_ifs):
        def get_kicad_footprint(self) -> str:
            obj = self.get_obj(SMDTwoPin)
            return "{typename}_SMD:{type}_{imperial}_{metric}Metric".format(
                typename=obj._type.name,
                type=obj._type.value,
                imperial=obj._size.imperial.without_prefix,
                metric=obj._size.metric.without_prefix,
            )

    kicad_footprint: _has_kicad_footprint
    equal_pins: F.has_equal_pins_in_ifs
    attach_via_pinmap: F.can_attach_via_pinmap_equal



from enum import IntEnum, auto

from faebryk.core.trait import TraitImpl
from faebryk.libs.library import L


class has_net_name(L.Trait.decless()):
    """Provide a net name suggestion or expectation"""

    # TODO:
    # Currently this is just a data-class, which is EXPECRTED to be used by
    # src/faebryk/exporters/netlist/graph.py to compute the net names
    # The intelligence of graph.py should be split and moved here

    class Level(IntEnum):
        SUGGESTED = auto()
        EXPECTED = auto()

    def __init__(self, name: str, level: Level = Level.SUGGESTED):
        super().__init__()
        self.name = name
        self.level = level

    @classmethod
    def suggested(cls, name: str) -> "has_net_name":
        return cls(name, cls.Level.SUGGESTED)

    @classmethod
    def expected(cls, name: str) -> "has_net_name":
        return cls(name, cls.Level.EXPECTED)

    def handle_duplicate(self, old: TraitImpl, node: L.Node) -> bool:
        assert isinstance(old, has_net_name)  # Asserting trait, not impl
        # FIXME: gracefully handle hitting this multiple times
        return super().handle_duplicate(old, node)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT
import logging

import faebryk.library._F as F
from faebryk.core.moduleinterface import ModuleInterface
from faebryk.libs.library import L
from faebryk.libs.util import times

logger = logging.getLogger(__name__)


class Addressor(ModuleInterface):
    address = L.p_field(domain=L.Domains.Numbers.NATURAL())
    offset = L.p_field(domain=L.Domains.Numbers.NATURAL())
    base = L.p_field(domain=L.Domains.Numbers.NATURAL())

    @L.rt_field
    def address_lines(self):
        return times(self._address_bits, F.ElectricLogic)

    @L.rt_field
    def single_electric_reference(self):
        return F.has_single_electric_reference_defined(
            F.ElectricLogic.connect_all_module_references(self)
        )

    def __init__(self, address_bits: int):
        self._address_bits = address_bits
        super().__init__()

    def __preinit__(self) -> None:
        for x in (self.address, self.offset, self.base):
            x.constrain_ge(0)

        self.offset.constrain_le(1 << self._address_bits)

        self.address.alias_is(self.base + self.offset)
        # TODO: not implemented yet
        # self.offset.constrain_cardinality(1)

        for i, line in enumerate(self.address_lines):
            (self.offset.operation_is_bit_set(i)).if_then_else(
                lambda line=line: line.set(True),
                lambda line=line: line.set(False),
            )



from unittest.mock import MagicMock

import pytest

from atopile.errors import (
    UserException,
)
from faebryk.libs.exceptions import (
    accumulate,
    downgrade,
    iter_leaf_exceptions,
    iter_through_errors,
    suppress_after_count,
)


def test_accumulate():
    with pytest.raises(UserException):
        with accumulate() as error_collector:
            with error_collector.collect():
                raise UserException("test error")

            raise UserException("test error 2")


def test_iter_through_errors():
    try:
        for cltr, i in iter_through_errors(range(4)):
            with cltr():
                if i == 1:
                    raise UserException("test error")
                if i == 2:
                    raise UserException("test error 2")

    except ExceptionGroup as ex:
        assert len(ex.exceptions) == 2
        ex_1, ex_2 = ex.exceptions
        assert ex_1.message == "test error"
        assert ex_2.message == "test error 2"

    else:
        raise AssertionError("Expected an ExceptionGroup to be raised")


def test_downgrade_context():
    logger = MagicMock()
    with downgrade(ValueError, logger=logger):
        raise ValueError()

    with pytest.raises(TypeError):
        with downgrade(ValueError, logger=logger):
            raise TypeError()

    logger.log.assert_called_once()


def test_downgrade_context_custom_exception():
    logger = MagicMock()

    class CustomException(Exception):
        pass

    with downgrade(CustomException, logger=logger):
        raise CustomException()

    with pytest.raises(Exception):
        with downgrade(CustomException, logger=logger):
            raise Exception()

    logger.log.assert_called_once()


def test_downgrade_context_multiple_exceptions():
    logger = MagicMock()
    with downgrade(ValueError, TypeError, logger=logger):
        raise ValueError()

    logger.log.assert_called_once()

    with pytest.raises(Exception):
        with downgrade(ValueError, TypeError, logger=logger):
            raise Exception()

    logger.log.assert_called_once()

    with downgrade(ValueError, TypeError, logger=logger):
        raise TypeError()


def test_downgrade_decorator():
    logger = MagicMock()

    @downgrade(ValueError, logger=logger)
    def foo():
        raise ValueError()

    a = foo()
    assert a is None
    logger.log.assert_called_once()


def test_downgrade_decorator_with_default():
    logger = MagicMock()

    @downgrade(ValueError, default=2, logger=logger)
    def foo():
        raise ValueError()

    a = foo()
    assert a == 2
    logger.log.assert_called_once()


def test_downgrade_raise_anyway():
    logger = MagicMock()
    with pytest.raises(ValueError):
        with downgrade(ValueError, raise_anyway=True, logger=logger):
            raise ValueError()


def test_suppress_after_count():
    logger = MagicMock()
    suppressor = suppress_after_count(
        3, ValueError, suppression_warning="test warning", logger=logger
    )
    for _ in range(3):
        with pytest.raises(ValueError), suppressor:
            raise ValueError()

    logger.warning.assert_not_called()

    with suppressor:
        raise ValueError()

    logger.warning.assert_called_once()


def test_iter_leaf_exceptions():
    ex = ExceptionGroup(
        "test",
        [
            Exception("test 1"),
            ExceptionGroup("test 2", [Exception("test 2.1"), Exception("test 2.2")]),
        ],
    )
    assert [ex.args[0] for ex in iter_leaf_exceptions(ex)] == [
        "test 1",
        "test 2.1",
        "test 2.2",
    ]



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.libs.library import L
from faebryk.libs.units import P


class SPIFlash(Module):
    power: F.ElectricPower
    qspi = L.f_field(F.MultiSPI)(4)

    memory_size = L.p_field(
        units=P.byte,
        domain=L.Domains.Numbers.NATURAL(),
    )
    designator_prefix = L.f_field(F.has_designator_prefix)(
        F.has_designator_prefix.Prefix.U
    )

    @L.rt_field
    def single_reference(self):
        return F.has_single_electric_reference_defined(
            F.ElectricLogic.connect_all_module_references(self)
        )



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
"""
Runner to use when running under a different interpreter.
"""

import os
import pathlib
import sys
import traceback


# **********************************************************
# Update sys.path before importing any bundled libraries.
# **********************************************************
def update_sys_path(path_to_add: str, strategy: str) -> None:
    """Add given path to `sys.path`."""
    if path_to_add not in sys.path and os.path.isdir(path_to_add):
        if strategy == "useBundled":
            sys.path.insert(0, path_to_add)
        elif strategy == "fromEnvironment":
            sys.path.append(path_to_add)


# Ensure that we can import LSP libraries, and other bundled libraries.
update_sys_path(
    os.fspath(pathlib.Path(__file__).parent.parent / "libs"),
    os.getenv("LS_IMPORT_STRATEGY", "useBundled"),
)


# pylint: disable=wrong-import-position,import-error
import atopile.lsp.lsp_jsonrpc as jsonrpc  # noqa: E402
import atopile.lsp.lsp_utils as utils  # noqa: E402


def run():
    RPC = jsonrpc.create_json_rpc(sys.stdin.buffer, sys.stdout.buffer)

    EXIT_NOW = False
    while not EXIT_NOW:
        msg = RPC.receive_data()

        method = msg["method"]
        if method == "exit":
            EXIT_NOW = True
            continue

        if method == "run":
            is_exception = False
            # This is needed to preserve sys.path, pylint modifies
            # sys.path and that might not work for this scenario
            # next time around.
            with utils.substitute_attr(sys, "path", sys.path[:]):
                # TODO: `utils.run_module` is equivalent to running`python -m atopile`.
                # If your tool supports a programmatic API then replace the function
                # below with code for your tool. You can also use `utils.run_api`
                # helper, which handles changing working directories, managing io
                # streams, etc. Also update `_run_tool_on_document` and `_run_tool`
                # functions in `lsp_server.py`.
                try:
                    result = utils.run_module(
                        module=msg["module"],
                        argv=msg["argv"],
                        use_stdin=msg["useStdin"],
                        cwd=msg["cwd"],
                        source=msg.get("source"),
                    )
                except Exception:  # pylint: disable=broad-except
                    result = utils.RunResult("", traceback.format_exc(chain=True))
                    is_exception = True

            response = {"id": msg["id"]}
            if result.stderr:
                response["error"] = result.stderr
                response["exception"] = is_exception
            elif result.stdout:
                response["result"] = result.stdout

            RPC.send_data(response)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

from abc import abstractmethod

from faebryk.core.moduleinterface import ModuleInterface
from faebryk.libs.util import cast_assert, find


class is_esphome_bus(ModuleInterface.TraitT):
    ...

    @abstractmethod
    def get_bus_id(self) -> str: ...

    @staticmethod
    def find_connected_bus[T: ModuleInterface](bus: T) -> T:
        connected_mifs = bus.get_connected()
        try:
            return cast_assert(
                type(bus),
                find(connected_mifs, lambda mif: mif.has_trait(is_esphome_bus)),
            )
        except ValueError:
            raise Exception(f"No esphome bus connected to {bus}: {connected_mifs}")



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
from abc import abstractmethod

from faebryk.core.moduleinterface import ModuleInterface

logger = logging.getLogger(__name__)


class is_optional(ModuleInterface.TraitT):
    """
    Indicates that a module might not have a specific interface
    """

    @abstractmethod
    def is_needed(self) -> bool:
        """
        Returns whether it's needed
        """
        pass

    @abstractmethod
    def _handle_result(self, needed: bool) -> None:
        """
        Handles the result of the is_needed method once it's known
        """
        pass



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

from faebryk.core.module import Module
from faebryk.libs.checksum import Checksum


class _FileManuallyModified(Exception): ...


class is_auto_generated(Module.TraitT.decless()):
    CHECKSUM_PLACEHOLDER = "{IS_AUTO_GENERATED_CHECKSUM}"

    def __init__(
        self,
        source: str | None = None,
        system: str | None = None,
        date: str | None = None,
        checksum: str | None = None,
    ) -> None:
        super().__init__()
        self._source = source
        self._system = system
        self._date = date
        self._checksum = checksum
        """
        checksum is the sha256 hash of the file where the checksum string is:
          '{CHECKSUM}'
        """

    @staticmethod
    def verify(stated_checksum: str, file_contents: str):
        with_placeholder = file_contents.replace(
            stated_checksum, is_auto_generated.CHECKSUM_PLACEHOLDER
        )
        try:
            Checksum.verify(stated_checksum, with_placeholder)
        except Checksum.Mismatch as e:
            raise _FileManuallyModified("File has been manually modified") from e

    @staticmethod
    def set_checksum(file_contents: str) -> str:
        actual = Checksum.build(file_contents)
        return file_contents.replace(is_auto_generated.CHECKSUM_PLACEHOLDER, actual)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import io
import logging
import os
import sys

import rich
from rich.console import Console
from rich.highlighter import RegexHighlighter
from rich.logging import RichHandler
from rich.markdown import Markdown
from rich.table import Table
from rich.theme import Theme
from rich.tree import Tree

from faebryk.libs.util import ConfigFlag, ConfigFlagInt


def rich_print_robust(message: str, markdown: bool = False):
    """
    Hack for terminals that don't support unicode
    There is probably a better way to do this, but this is a quick fix for now.
    """
    try:
        rich.print(Markdown(message) if markdown else message)
    except UnicodeEncodeError:
        message = message.encode("ascii", errors="ignore").decode("ascii")
        rich.print(Markdown(message) if markdown else message)


def is_piped_to_file() -> bool:
    return not sys.stdout.isatty()


def get_terminal_width() -> int:
    if is_piped_to_file():
        if "COLUMNS" in os.environ:
            return int(os.environ["COLUMNS"])
        else:
            return 240
    else:
        return Console().size.width


PLOG = ConfigFlag("PLOG", descr="Enable picker debug log")
FLOG_FMT = ConfigFlag("LOG_FMT", descr="Enable (old) log formatting")
TERMINAL_WIDTH = ConfigFlagInt(
    "TERMINAL_WIDTH",
    default=get_terminal_width(),
    descr="Width of the terminal",
)
NET_LINE_WIDTH = int(TERMINAL_WIDTH) - 40


class NestedConsole(Console):
    def __init__(self, *args, **kwargs):
        super().__init__(
            *args, record=True, width=NET_LINE_WIDTH, file=io.StringIO(), **kwargs
        )

    def __str__(self):
        return self.export_text(styles=True)


def rich_to_string(rich_obj: Table | Tree) -> str:
    console = NestedConsole()
    console.print(rich_obj)
    return str(console)


def setup_basic_logging():
    if FLOG_FMT:
        logging.basicConfig(
            format="%(message)s",
            level=logging.INFO,
            datefmt="[%H:%M:%S]",
            handlers=[
                RichHandler(
                    console=Console(
                        safe_box=False,
                        theme=theme,
                        force_terminal=True,
                        width=int(TERMINAL_WIDTH),
                    ),
                    highlighter=NodeHighlighter(),
                )
            ],
        )

    if PLOG:
        from faebryk.libs.picker.picker import logger as plog

        plog.setLevel(logging.DEBUG)


class NodeHighlighter(RegexHighlighter):
    """
    Apply style to anything that looks like an faebryk Node\n
    <*|XOR_with_NANDS.nands[2]|NAND.inputs[0]|Logic> with
      <*|TI_CD4011BE.nands[2]|ElectricNAND.inputs[0]|ElectricLogic>\n
    \t<> = Node\n
    \t|  = Type\n
    \t.  = Parent\n
    \t*  = Root
    """

    base_style = "node."
    highlights = [
        #  r"(?P<Rest>(.*))",
        r"(?P<Node>([/</>]))",
        r"[?=\|](?P<Type>([a-zA-Z_0-9]+))[?=\>]",
        r"[\.](?P<Child>([a-zA-Z_0-9]+))[?=\[]",
        r"[\|](?P<Parent>([a-zA-Z_0-9]+))[?=\.]",
        r"[?<=*.](?P<Root>(\*))",
        r"[?=\[](?P<Number>([0-9]+))[?=\]]",
        # Solver/Parameter stuff -------------------------------------------------------
        # Literals
        r"(?P<Quantity>Quantity_Interval(_Disjoint)?\([^)]*\))",
        r"(?P<Quantity>\(\[[^)]*\]\))",
        r"(?P<Quantity>\[(True|False)+\])",
        # Predicates / Expressions
        r"(?P<Op> (\+|\*|/))[ {]",
        r"(?P<Predicate> (is|⊆|≥|≤|)!?!?[✓✗]?) ",
        # Literal Is/IsSubset
        r"(?P<IsSubset>{(I|S)\|[^}]+})",
    ]


class ReprHighlighter(RegexHighlighter):
    """Highlights the text typically produced from ``__repr__`` methods."""

    base_style = "repr."
    highlights = [
        r"(?P<tag_start><)(?P<tag_name>[-\w.:|]*)(?P<tag_contents>[\w\W]*)(?P<tag_end>>)",
        r'(?P<attrib_name>[\w_]{1,50})=(?P<attrib_value>"?[\w_]+"?)?',
        r"(?P<brace>[][{}()])",
        "|".join(
            [
                r"(?P<ipv4>[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3})",
                r"(?P<ipv6>([A-Fa-f0-9]{1,4}::?){1,7}[A-Fa-f0-9]{1,4})",
                r"(?P<eui64>(?:[0-9A-Fa-f]{1,2}-){7}[0-9A-Fa-f]{1,2}|(?:[0-9A-Fa-f]{1,2}:){7}[0-9A-Fa-f]{1,2}|(?:[0-9A-Fa-f]{4}\.){3}[0-9A-Fa-f]{4})",
                r"(?P<eui48>(?:[0-9A-Fa-f]{1,2}-){5}[0-9A-Fa-f]{1,2}|(?:[0-9A-Fa-f]{1,2}:){5}[0-9A-Fa-f]{1,2}|(?:[0-9A-Fa-f]{4}\.){2}[0-9A-Fa-f]{4})",
                r"(?P<call>[\w.]*?)\(",
                r"\b(?P<bool_true>True)\b|\b(?P<bool_false>False)\b|\b(?P<none>None)\b",
                r"(?P<ellipsis>\.\.\.)",
                r"(?P<number_complex>(?<!\w)(?:\-?[0-9]+\.?[0-9]*(?:e[-+]?\d+?)?)(?:[-+](?:[0-9]+\.?[0-9]*(?:e[-+]?\d+)?))?j)",
                r"(?P<number>(?<!\w)\-?[0-9]+\.?[0-9]*(e[-+]?\d+?)?\b|0x[0-9a-fA-F]*)",
                # matches paths with or without leading / (@windows)
                r"(?P<path>(\b|/)([-\w._+]+)(/[-\w._+]+)*/)(?P<filename>[-\w._+]*)?",
                r"(?<![\\\w])(?P<str>b?'''.*?(?<!\\)'''|b?'.*?(?<!\\)'|b?\"\"\".*?(?<!\\)\"\"\"|b?\".*?(?<!\\)\")",
                r"(?P<url>(file|https|http|ws|wss)://[-0-9a-zA-Z$_+!`(),.?/;:&=%#~@]*)",
            ]
        ),
    ]


theme = Theme(
    {
        "node.Node": "bold magenta",
        "node.Type": "bright_cyan",
        "node.Parent": "bright_red",
        "node.Child": "bright_yellow",
        "node.Root": "bold yellow",
        "node.Number": "bright_green",
        #   "node.Rest": "bright_black",
        "logging.level.warning": "yellow",
        "node.Quantity": "bright_yellow",
        "node.IsSubset": "bright_blue",
        "node.Predicate": "bright_magenta",
        "node.Op": "red",
    }
)



"""
We collect anonymous telemetry data to help improve atopile.
To opt out, add `telemetry: false` to your project's ~/atopile/telemetry.yaml file.

What we collect:
- Hashed user id so we know how many unique users we have
- Hashed project id
- Error logs
- How long the build took
- ato version
- Git hash of current commit
"""

import contextlib
import hashlib
import importlib.metadata
import logging
import os
import time
import uuid
from collections.abc import Generator
from contextlib import contextmanager
from dataclasses import asdict, dataclass, field
from typing import Any, Unpack

from posthog import Posthog
from posthog.args import OptionalCaptureArgs
from ruamel.yaml import YAML

from faebryk.libs.paths import get_config_dir
from faebryk.libs.util import cast_assert, once

log = logging.getLogger(__name__)


class _MockClient:
    disabled = False

    def capture_exception(
        self,
        exception: Exception,
        **kwargs: Unpack[OptionalCaptureArgs],
    ) -> None:
        pass

    def capture(
        self,
        event: str,
        **kwargs: Unpack[OptionalCaptureArgs],
    ) -> None:
        pass


@once
def _get_posthog_client() -> Posthog | _MockClient:
    try:
        return Posthog(
            # write-only API key, intended to be made public
            project_api_key="phc_IIl9Bip0fvyIzQFaOAubMYYM2aNZcn26Y784HcTeMVt",
            host="https://telemetry.atopileapi.com",
            sync_mode=True,
        )
    except Exception as e:
        log.debug("Failed to initialize telemetry client: %s", e, exc_info=e)
        return _MockClient()


client = _get_posthog_client()


@dataclass
class TelemetryConfig:
    telemetry: bool | None = True
    id: uuid.UUID | None = field(default_factory=uuid.uuid4)

    @classmethod
    @once
    def load(cls) -> "TelemetryConfig":
        atopile_config_dir = get_config_dir()
        atopile_yaml = atopile_config_dir / "telemetry.yaml"

        if not atopile_yaml.exists():
            config = TelemetryConfig()
            atopile_config_dir.mkdir(parents=True, exist_ok=True)
            with atopile_yaml.open("w", encoding="utf-8") as f:
                yaml = YAML()
                yaml.dump(config, f)
            return config

        with atopile_yaml.open(encoding="utf-8") as f:
            yaml = YAML()
            config = TelemetryConfig(**yaml.load(f))

        if config.telemetry is False:
            log.log(0, "Telemetry is disabled. Skipping telemetry logging.")
            client.disabled = True

        return config


def _normalize_git_remote_url(git_url: str) -> str:
    """
    Commonize the remote which could be in either of these forms:
        - https://github.com/atopile/atopile.git
        - git@github.com:atopile/atopile.git
    ... to "github.com/atopile/atopile"
    """

    if git_url.startswith("git@"):
        git_url = git_url.removeprefix("git@")
        git_url = "/".join(git_url.split(":", 1))
    else:
        git_url = git_url.removeprefix("https://")

    git_url = git_url.removesuffix(".git")

    return git_url


class PropertyLoaders:
    @once
    @staticmethod
    def email() -> str | None:
        """Get the git user email."""
        try:
            import git
        except ImportError:
            return None

        with contextlib.suppress(
            git.InvalidGitRepositoryError,
            git.NoSuchPathError,
            ValueError,
            AttributeError,
        ):
            repo = git.Repo(search_parent_directories=True)
            config_reader = repo.config_reader()
            return cast_assert(str, config_reader.get_value("user", "email", None))

    @once
    @staticmethod
    def current_git_hash() -> str | None:
        """Get the current git commit hash."""
        try:
            import git
        except ImportError:
            return None

        with contextlib.suppress(
            git.InvalidGitRepositoryError,
            git.NoSuchPathError,
            ValueError,
            AttributeError,
        ):
            repo = git.Repo(search_parent_directories=True)
            return repo.head.commit.hexsha

    @once
    @staticmethod
    def project_id() -> str | None:
        """Get the hashed project ID from the git URL of the project, if available."""
        try:
            import git
        except ImportError:
            # no git executable
            return None

        try:
            repo = git.Repo(search_parent_directories=True)
        except (
            git.InvalidGitRepositoryError,
            git.NoSuchPathError,
            ValueError,
            AttributeError,
        ):
            return None

        if not repo.remotes:
            return None

        if (git_url := repo.remotes.origin.url) is None:
            return None

        project_url = _normalize_git_remote_url(git_url)

        log.log(0, "Project URL: %s", project_url)

        # Hash the project ID to de-identify it
        return hashlib.sha256(project_url.encode()).hexdigest()

    @once
    @staticmethod
    def ci_provider() -> str | None:
        if os.getenv("GITHUB_ACTIONS"):
            return "GitHub Actions"
        elif os.getenv("TF_BUILD"):
            return "Azure Pipelines"
        elif os.getenv("CIRCLECI"):
            return "Circle CI"
        elif os.getenv("TRAVIS"):
            return "Travis CI"
        elif os.getenv("BUILDKITE"):
            return "Buildkite"
        elif os.getenv("CIRRUS_CI"):
            return "Cirrus CI"
        elif os.getenv("GITLAB_CI"):
            return "GitLab CI"
        elif os.getenv("TEAMCITY_VERSION"):
            return "TeamCity"
        elif os.getenv("CODEBUILD_BUILD_ID"):
            return "CodeBuild"
        elif os.getenv("HEROKU_TEST_RUN_ID"):
            return "Heroku CI"
        elif os.getenv("bamboo.buildKey"):
            return "Bamboo"
        elif os.getenv("BUILD_ID"):
            return "Jenkins"  # could also be Hudson
        elif os.getenv("CI"):
            return "Other"

        return None


@dataclass
class TelemetryProperties:
    duration: float | None = None
    email: str | None = field(default_factory=PropertyLoaders.email)
    current_git_hash: str | None = field(
        default_factory=PropertyLoaders.current_git_hash
    )
    project_id: str | None = field(default_factory=PropertyLoaders.project_id)
    ci_provider: str | None = field(default_factory=PropertyLoaders.ci_provider)
    atopile_version: str = field(
        default_factory=lambda: importlib.metadata.version("atopile")
    )

    def __post_init__(self) -> None:
        self._start_time = time.perf_counter()

    def prepare(self, properties: dict | None = None) -> dict:
        now = time.perf_counter()
        self.duration = now - (self._start_time or now)
        return {**asdict(self), **(properties or {})}


def capture_exception(exc: Exception, properties: dict | None = None) -> None:
    try:
        config = TelemetryConfig.load()
    except Exception as e:
        log.debug("Failed to load telemetry config: %s", e, exc_info=e)
        return

    if config.telemetry is False:
        return

    default_properties = TelemetryProperties()
    properties = default_properties.prepare(properties)

    try:
        client.capture_exception(exc, distinct_id=config.id, properties=properties)
    except Exception as e:
        log.debug("Failed to send exception telemetry data: %s", e, exc_info=e)


@contextmanager
def capture(
    event_start: str, event_end: str, properties: dict | None = None
) -> Generator[None, Any, None]:
    try:
        config = TelemetryConfig.load()
    except Exception as e:
        log.debug("Failed to load telemetry config: %s", e, exc_info=e)
        yield
        return

    if config.telemetry is False:
        yield
        return

    default_properties = TelemetryProperties()

    try:
        client.capture(
            distinct_id=config.id,
            event=event_start,
            properties=default_properties.prepare(properties),
        )
    except Exception as e:
        log.debug("Failed to send telemetry data (event start): %s", e, exc_info=e)
        yield
        return

    try:
        yield
    except Exception as e:
        try:
            client.capture_exception(
                e,
                distinct_id=config.id,
                properties=default_properties.prepare(properties),
            )
        except Exception as e:
            log.debug("Failed to send exception telemetry data: %s", e, exc_info=e)
        raise

    try:
        client.capture(
            distinct_id=config.id,
            event=event_end,
            properties=default_properties.prepare(properties),
        )
    except Exception as e:
        log.debug("Failed to send telemetry data (event end): %s", e, exc_info=e)
